@@ -1,7 +1,7 @@
 #!/bin/bash
 # StarPU --- Runtime system for heterogeneous multicore architectures.
 #
-# Copyright (C) 2018                 Université de Bordeaux
+# Copyright (C) 2018                                     Université de Bordeaux
 #
 # StarPU is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -53,10 +53,12 @@ plot \\
 EOF
 
 N=2
+COMMA=""
 for STARPU_SCHED in $STARPU_SCHEDS
 do
-	echo "'cholesky.output' using 1:$N with lines title '$STARPU_SCHED', \\" >> cholesky.gp
+	echo "$COMMA'cholesky.output' using 1:$N with lines title '$STARPU_SCHED' \\" >> cholesky.gp
 	N=$(($N + 1))
+	COMMA=", "
 done
 gnuplot cholesky.gp
 gv $OUTFILE
