@@ -1,7 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009, 2010, 2014  Université de Bordeaux 1
- * Copyright (C) 2010  Centre National de la Recherche Scientifique
+ * Copyright (C) 2009-2011,2014,2015,2018                 Université de Bordeaux
+ * Copyright (C) 2012                                     Inria
+ * Copyright (C) 2010,2015,2017                           CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -238,7 +239,7 @@ void STARPU_DSWAP(const int n, double *x, const int incx, double *y, const int i
 	cblas_dswap(n, x, incx, y, incy);
 }
 
-#elif defined(STARPU_GOTO) || defined(STARPU_SYSTEM_BLAS) || defined(STARPU_MKL)
+#elif defined(STARPU_GOTO) || defined(STARPU_OPENBLAS) || defined(STARPU_SYSTEM_BLAS) || defined(STARPU_MKL) || defined(STARPU_ARMPL)
 
 inline void STARPU_SGEMM(char *transa, char *transb, int M, int N, int K, 
 			float alpha, const float *A, int lda, const float *B, int ldb, 
@@ -414,6 +415,19 @@ void STARPU_DSWAP(const int n, double *X, const int incX, double *Y, const int i
 	dswap_(&n, X, &incX, Y, &incY);
 }
 
+#if defined(STARPU_MKL) || defined(STARPU_ARMPL)
+void STARPU_SPOTRF(const char*uplo, const int n, float *a, const int lda)
+{
+	int info = 0;
+	spotrf_(uplo, &n, a, &lda, &info);
+}
+
+void STARPU_DPOTRF(const char*uplo, const int n, double *a, const int lda)
+{
+	int info = 0;
+	dpotrf_(uplo, &n, a, &lda, &info);
+}
+#endif
 
 #elif defined(STARPU_SIMGRID)
 inline void STARPU_SGEMM(char *transa, char *transb, int M, int N, int K, 
@@ -430,9 +444,9 @@ inline void STARPU_SGEMV(char *transa, int M, int N, float alpha, float *A, int
 inline void STARPU_DGEMV(char *transa, int M, int N, double alpha, double *A, int lda,
 		double *X, int incX, double beta, double *Y, int incY) { }
 
-inline float STARPU_SASUM(int N, float *X, int incX) { }
+inline float STARPU_SASUM(int N, float *X, int incX) { return 0.; }
 
-inline double STARPU_DASUM(int N, double *X, int incX) { }
+inline double STARPU_DASUM(int N, double *X, int incX) { return 0.; }
 
 void STARPU_SSCAL(int N, float alpha, float *X, int incX) { }
 
@@ -486,18 +500,22 @@ void STARPU_SAXPY(const int n, const float alpha, float *X, const int incX, floa
 
 void STARPU_DAXPY(const int n, const double alpha, double *X, const int incX, double *Y, const int incY) { }
 
-int STARPU_ISAMAX (const int n, float *X, const int incX) { }
+int STARPU_ISAMAX (const int n, float *X, const int incX) { return 0; }
 
-int STARPU_IDAMAX (const int n, double *X, const int incX) { }
+int STARPU_IDAMAX (const int n, double *X, const int incX) { return 0; }
 
-float STARPU_SDOT(const int n, const float *x, const int incx, const float *y, const int incy) { }
+float STARPU_SDOT(const int n, const float *x, const int incx, const float *y, const int incy) { return 0.; }
 
-double STARPU_DDOT(const int n, const double *x, const int incx, const double *y, const int incy) { }
+double STARPU_DDOT(const int n, const double *x, const int incx, const double *y, const int incy) { return 0.; }
 
 void STARPU_SSWAP(const int n, float *X, const int incX, float *Y, const int incY) { }
 
 void STARPU_DSWAP(const int n, double *X, const int incX, double *Y, const int incY) { }
 
+void STARPU_SPOTRF(const char*uplo, const int n, float *a, const int lda) { }
+
+void STARPU_DPOTRF(const char*uplo, const int n, double *a, const int lda) { }
+
 
 #else
 #error "no BLAS lib available..."
