@@ -2,6 +2,7 @@
 # StarPU --- Runtime system for heterogeneous multicore architectures.
 #
 # Copyright (C) 2017                                     CNRS
+# Copyright (C) 2017                                     Inria
 # Copyright (C) 2017                                     Université de Bordeaux
 #
 # StarPU is free software; you can redistribute it and/or modify
@@ -21,6 +22,12 @@ set -e
 
 PREFIX=$(dirname $0)
 
+if [ -n "$STARPU_MIC_SINK_PROGRAM_PATH" ] ; then
+	STARPU_MIC_SINK_PROGRAM_NAME=$STARPU_MIC_SINK_PROGRAM_PATH/heat
+	# in case libtool got into play
+	[ -x "$STARPU_MIC_SINK_PROGRAM_PATH/.libs/heat" ] && STARPU_MIC_SINK_PROGRAM_NAME=$STARPU_MIC_SINK_PROGRAM_PATH/.libs/heat
+fi
+
 $PREFIX/heat -shape 0
 $PREFIX/heat -shape 1
 # sometimes lead to pivot being 0
