@@ -1,8 +1,8 @@
 #!/bin/sh
-
 # StarPU --- Runtime system for heterogeneous multicore architectures.
 #
-# Copyright (C) 2012, 2014  CNRS
+# Copyright (C) 2012,2014,2015,2017,2018                 CNRS
+# Copyright (C) 2017,2019                                Université de Bordeaux
 #
 # StarPU is free software; you can redistribute it and/or modify
 # it under the terms of the GNU Lesser General Public License as published by
@@ -14,7 +14,7 @@
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 #
 # See the GNU Lesser General Public License in COPYING.LGPL for more details.
-
+#
 check_success()
 {
     if [ $1 -ne 0 ] ; then
@@ -29,7 +29,7 @@ then
     exit 77
 fi
 
-SCHEDULERS=`STARPU_SCHED="help" ./basic_examples/hello_world 2>&1 | awk '/\t->/ {print $1}'`
+SCHEDULERS=`../tools/starpu_sched_display | grep -v pheft | grep -v peager | grep -v heteroprio | grep -v modular-gemm`
 
 for sched in $SCHEDULERS
 do
