@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
  * Copyright (C) 2017                                     Inria
- * Copyright (C) 2017-2018                                CNRS
- * Copyright (C) 2017-2018                                Université de Bordeaux
+ * Copyright (C) 2017,2018                                CNRS
+ * Copyright (C) 2017,2018                                Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -20,15 +20,12 @@
 #include "helper.h"
 #include <inttypes.h>
 
-int main(int argc, char **argv)
+int main(int argc STARPU_ATTRIBUTE_UNUSED, char *argv[] STARPU_ATTRIBUTE_UNUSED)
 {
 	int flag;
 	int64_t *value;
 	int64_t rvalue;
 
-	(void) argc;
-	(void) argv;
-
 	starpu_mpi_comm_get_attr(MPI_COMM_WORLD, 42, NULL, &flag);
 	STARPU_ASSERT_MSG(flag == 0, "starpu_mpi_comm_get_attr was called with invalid argument\n");
 
