@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2011-2013                                Inria
- * Copyright (C) 2008-2017                                Université de Bordeaux
- * Copyright (C) 2010-2017                                CNRS
+ * Copyright (C) 2011-2014,2016,2017                      Inria
+ * Copyright (C) 2008-2019                                Université de Bordeaux
+ * Copyright (C) 2010-2017, 2019                          CNRS
  * Copyright (C) 2013                                     Thibaut Lambert
  * Copyright (C) 2011                                     Télécom-SudParis
  *
@@ -33,10 +33,12 @@
 #include <core/workers.h>
 #include <datawizard/datawizard.h>
 #include <core/perfmodel/regression.h>
+#include <core/perfmodel/multiple_regression.h>
 #include <common/config.h>
 #include <starpu_parameters.h>
 #include <common/uthash.h>
 #include <limits.h>
+#include <core/task.h>
 
 #ifdef STARPU_HAVE_WINDOWS
 #include <windows.h>
@@ -91,7 +93,13 @@ void _starpu_perfmodel_malloc_per_arch(struct starpu_perfmodel *model, int comb,
 
 void _starpu_perfmodel_malloc_per_arch_is_set(struct starpu_perfmodel *model, int comb, int nb_impl)
 {
-	_STARPU_CALLOC(model->state->per_arch_is_set[comb], nb_impl, sizeof(int));
+	int i;
+
+	_STARPU_MALLOC(model->state->per_arch_is_set[comb], nb_impl*sizeof(int));
+	for(i = 0; i < nb_impl; i++)
+	{
+		model->state->per_arch_is_set[comb][i] = 0;
+	}
 }
 
 int _starpu_perfmodel_arch_comb_get(int ndevices, struct starpu_perfmodel_device *devices)
@@ -242,6 +250,64 @@ static void insert_history_entry(struct starpu_perfmodel_history_entry *entry, s
 }
 
 #ifndef STARPU_SIMGRID
+static void check_reg_model(struct starpu_perfmodel *model, int comb, int impl)
+{
+	struct starpu_perfmodel_per_arch *per_arch_model = &model->state->per_arch[comb][impl];
+	struct starpu_perfmodel_regression_model *reg_model = &per_arch_model->regression;
+
+	/*
+	 * Linear Regression model
+	 */
+
+	/* Unless we have enough measurements, we put NaN in the file to indicate the model is invalid */
+	double alpha = nan(""), beta = nan("");
+	if (model->type == STARPU_REGRESSION_BASED || model->type == STARPU_NL_REGRESSION_BASED)
+	{
+		if (reg_model->nsample > 1)
+		{
+			alpha = reg_model->alpha;
+			beta = reg_model->beta;
+		}
+	}
+
+	/* TODO: check:
+	 * reg_model->sumlnx
+	 * reg_model->sumlnx2
+	 * reg_model->sumlny
+	 * reg_model->sumlnxlny
+	 * alpha
+	 * beta
+	 * reg_model->minx
+	 * reg_model->maxx
+	 */
+	(void)alpha;
+	(void)beta;
+
+	/*
+	 * Non-Linear Regression model
+	 */
+
+	double a = nan(""), b = nan(""), c = nan("");
+
+	if (model->type == STARPU_NL_REGRESSION_BASED)
+		_starpu_regression_non_linear_power(per_arch_model->list, &a, &b, &c);
+
+	/* TODO: check:
+	 * a
+	 * b
+	 * c
+	 */
+
+	/*
+	 * Multiple Regression Model
+	 */
+
+	if (model->type == STARPU_MULTIPLE_REGRESSION_BASED)
+	{
+		/* TODO: check: */
+	}
+}
+
 static void dump_reg_model(FILE *f, struct starpu_perfmodel *model, int comb, int impl)
 {
 	struct starpu_perfmodel_per_arch *per_arch_model;
@@ -288,6 +354,67 @@ static void dump_reg_model(FILE *f, struct starpu_perfmodel *model, int comb, in
 	fprintf(f, "\t");
 	_starpu_write_double(f, "%-15e", c);
 	fprintf(f, "\n");
+
+	/*
+	 * Multiple Regression Model
+	 */
+
+	if (model->type != STARPU_MULTIPLE_REGRESSION_BASED)
+	{
+		fprintf(f, "# not multiple-regression-base\n");
+		fprintf(f, "0\n");
+	}
+	else
+	{
+		if (reg_model->ncoeff==0 && model->ncombinations!=0 && model->combinations!=NULL)
+		{
+			reg_model->ncoeff = model->ncombinations + 1;
+		}
+
+		_STARPU_MALLOC(reg_model->coeff,  reg_model->ncoeff*sizeof(double));
+		_starpu_multiple_regression(per_arch_model->list, reg_model->coeff, reg_model->ncoeff, model->nparameters, model->parameters_names, model->combinations, model->symbol);
+
+		fprintf(f, "# n\tintercept\t");
+		if (reg_model->ncoeff==0 || model->ncombinations==0 || model->combinations==NULL)
+			fprintf(f, "\n1\tnan");
+		else
+		{
+			unsigned i;
+			for (i=0; i < model->ncombinations; i++)
+			{
+				if (model->parameters_names == NULL)
+					fprintf(f, "c%u", i+1);
+				else
+				{
+					unsigned j;
+					int first=1;
+					for(j=0; j < model->nparameters; j++)
+					{
+						if (model->combinations[i][j] > 0)
+						{
+							if (first)
+								first=0;
+							else
+								fprintf(f, "*");
+
+							if(model->parameters_names[j] != NULL)
+								fprintf(f, "%s", model->parameters_names[j]);
+							else
+								fprintf(f, "P%u", j);
+
+							if (model->combinations[i][j] > 1)
+								fprintf(f, "^%d", model->combinations[i][j]);
+						}
+					}
+				}
+				fprintf(f, "\t\t");
+			}
+
+			fprintf(f, "\n%u", reg_model->ncoeff);
+			for (i=0; i < reg_model->ncoeff; i++)
+				fprintf(f, "\t%-15e", reg_model->coeff[i]);
+		}
+	}
 }
 #endif
 
@@ -332,9 +459,45 @@ static void scan_reg_model(FILE *f, const char *path, struct starpu_perfmodel_re
 	/* If any of the parameters describing the non-linear regression model is NaN, the model is invalid */
 	unsigned nl_invalid = (isnan(reg_model->a)||isnan(reg_model->b)||isnan(reg_model->c));
 	reg_model->nl_valid = !nl_invalid && VALID_REGRESSION(reg_model);
+
+	_starpu_drop_comments(f);
+
+	// Read how many coefficients is there
+	res = fscanf(f, "%u", &reg_model->ncoeff);
+	STARPU_ASSERT_MSG(res == 1, "Incorrect performance model file %s", path);
+
+	/*
+	 * Multiple Regression Model
+	 */
+	if (reg_model->ncoeff != 0)
+	{
+		_STARPU_MALLOC(reg_model->coeff, reg_model->ncoeff*sizeof(double));
+
+		unsigned multi_invalid = 0;
+		unsigned i;
+		for (i=0; i < reg_model->ncoeff; i++)
+		{
+			res = _starpu_read_double(f, "%le", &reg_model->coeff[i]);
+			STARPU_ASSERT_MSG(res == 1, "Incorrect performance model file %s", path);
+			multi_invalid = (multi_invalid||isnan(reg_model->coeff[i]));
+		}
+		reg_model->multi_valid = !multi_invalid;
+	}
+	res = fscanf(f, "\n");
+	STARPU_ASSERT_MSG(res == 0, "Incorrect performance model file %s", path);
 }
 
+
 #ifndef STARPU_SIMGRID
+static void check_history_entry(struct starpu_perfmodel_history_entry *entry)
+{
+	STARPU_ASSERT_MSG(entry->deviation >= 0, "entry=%p, entry->deviation=%lf\n", entry, entry->deviation);
+	STARPU_ASSERT_MSG(entry->sum >= 0, "entry=%p, entry->sum=%lf\n", entry, entry->sum);
+	STARPU_ASSERT_MSG(entry->sum2 >= 0, "entry=%p, entry->sum2=%lf\n", entry, entry->sum2);
+	STARPU_ASSERT_MSG(entry->mean >= 0, "entry=%p, entry->mean=%lf\n", entry, entry->mean);
+	STARPU_ASSERT_MSG(isnan(entry->flops)||entry->flops >= 0, "entry=%p, entry->flops=%lf\n", entry, entry->flops);
+	STARPU_ASSERT_MSG(entry->duration >= 0, "entry=%p, entry->duration=%lf\n", entry, entry->duration);
+}
 static void dump_history_entry(FILE *f, struct starpu_perfmodel_history_entry *entry)
 {
 	fprintf(f, "%08x\t%-15lu\t%-15e\t%-15e\t%-15e\t%-15e\t%-15e\t%u\n", entry->footprint, (unsigned long) entry->size, entry->flops, entry->mean, entry->deviation, entry->sum, entry->sum2, entry->nsample);
@@ -377,6 +540,11 @@ static void scan_history_entry(FILE *f, const char *path, struct starpu_perfmode
 
 	if (entry)
 	{
+		STARPU_ASSERT_MSG(flops >=0, "Negative flops %lf in performance model file %s", flops, path);
+		STARPU_ASSERT_MSG(mean >=0, "Negative mean %lf in performance model file %s", mean, path);
+		STARPU_ASSERT_MSG(deviation >=0, "Negative deviation %lf in performance model file %s", deviation, path);
+		STARPU_ASSERT_MSG(sum >=0, "Negative sum %lf in performance model file %s", sum, path);
+		STARPU_ASSERT_MSG(sum2 >=0, "Negative sum2 %lf in performance model file %s", sum2, path);
 		entry->footprint = footprint;
 		entry->size = size;
 		entry->flops = flops;
@@ -388,16 +556,17 @@ static void scan_history_entry(FILE *f, const char *path, struct starpu_perfmode
 	}
 }
 
-static void parse_per_arch_model_file(FILE *f, const char *path, struct starpu_perfmodel_per_arch *per_arch_model, unsigned scan_history)
+static void parse_per_arch_model_file(FILE *f, const char *path, struct starpu_perfmodel_per_arch *per_arch_model, unsigned scan_history, struct starpu_perfmodel *model)
 {
 	unsigned nentries;
+	struct starpu_perfmodel_regression_model *reg_model = &per_arch_model->regression;
 
 	_starpu_drop_comments(f);
 
 	int res = fscanf(f, "%u\n", &nentries);
 	STARPU_ASSERT_MSG(res == 1, "Incorrect performance model file %s", path);
 
-	scan_reg_model(f, path, &per_arch_model->regression);
+	scan_reg_model(f, path, reg_model);
 
 	/* parse entries */
 	unsigned i;
@@ -406,7 +575,7 @@ static void parse_per_arch_model_file(FILE *f, const char *path, struct starpu_p
 		struct starpu_perfmodel_history_entry *entry = NULL;
 		if (scan_history)
 		{
-			_STARPU_MALLOC(entry, sizeof(struct starpu_perfmodel_history_entry));
+			_STARPU_CALLOC(entry, 1, sizeof(struct starpu_perfmodel_history_entry));
 
 			/* Tell  helgrind that we do not care about
 			 * racing access to the sampling, we only want a
@@ -424,13 +593,27 @@ static void parse_per_arch_model_file(FILE *f, const char *path, struct starpu_p
 		if (scan_history)
 			insert_history_entry(entry, &per_arch_model->list, &per_arch_model->history);
 	}
+
+	if (model && model->type == STARPU_PERFMODEL_INVALID)
+	{
+		/* Tool loading a perfmodel without having the corresponding codelet */
+		if (reg_model->ncoeff != 0)
+			model->type = STARPU_MULTIPLE_REGRESSION_BASED;
+		else if (!isnan(reg_model->a) && !isnan(reg_model->b) && !isnan(reg_model->c))
+			model->type = STARPU_NL_REGRESSION_BASED;
+		else if (!isnan(reg_model->alpha) && !isnan(reg_model->beta))
+			model->type = STARPU_REGRESSION_BASED;
+		else if (nentries)
+			model->type = STARPU_HISTORY_BASED;
+		/* else unknown, leave invalid */
+	}
 }
 
 
 static void parse_arch(FILE *f, const char *path, struct starpu_perfmodel *model, unsigned scan_history, int comb)
 {
 	struct starpu_perfmodel_per_arch dummy;
-	unsigned nimpls, implmax, impl, i, ret;
+	unsigned nimpls, impl, i, ret;
 
 	/* Parsing number of implementation */
 	_starpu_drop_comments(f);
@@ -440,7 +623,7 @@ static void parse_arch(FILE *f, const char *path, struct starpu_perfmodel *model
 	if( model != NULL)
 	{
 		/* Parsing each implementation */
-		implmax = STARPU_MIN(nimpls, STARPU_MAXIMPLEMENTATIONS);
+		unsigned implmax = STARPU_MIN(nimpls, STARPU_MAXIMPLEMENTATIONS);
 		model->state->nimpls[comb] = implmax;
 		if (!model->state->per_arch[comb])
 		{
@@ -455,7 +638,7 @@ static void parse_arch(FILE *f, const char *path, struct starpu_perfmodel *model
 		{
 			struct starpu_perfmodel_per_arch *per_arch_model = &model->state->per_arch[comb][impl];
 			model->state->per_arch_is_set[comb][impl] = 1;
-			parse_per_arch_model_file(f, path, per_arch_model, scan_history);
+			parse_per_arch_model_file(f, path, per_arch_model, scan_history, model);
 		}
 	}
 	else
@@ -466,28 +649,7 @@ static void parse_arch(FILE *f, const char *path, struct starpu_perfmodel *model
 	/* if the number of implementation is greater than STARPU_MAXIMPLEMENTATIONS
 	 * we skip the last implementation */
 	for (i = impl; i < nimpls; i++)
-		parse_per_arch_model_file(f, path, &dummy, 0);
-
-}
-
-static enum starpu_worker_archtype _get_enum_type(int type)
-{
-	switch(type)
-	{
-	        case 0:
-			return STARPU_CPU_WORKER;
-        	case 1:
-			return STARPU_CUDA_WORKER;
-	        case 2:
-			return STARPU_OPENCL_WORKER;
-        	case 3:
-			return STARPU_MIC_WORKER;
-        	case 4:
-			return STARPU_SCC_WORKER;
-		default:
-			STARPU_ABORT();
-	}
-
+		parse_per_arch_model_file(f, path, &dummy, 0, NULL);
 }
 
 static void parse_comb(FILE *f, const char *path, struct starpu_perfmodel *model, unsigned scan_history, int comb)
@@ -502,12 +664,10 @@ static void parse_comb(FILE *f, const char *path, struct starpu_perfmodel *model
 	int dev;
 	for(dev = 0; dev < ndevices; dev++)
 	{
-		enum starpu_worker_archtype dev_type;
 		_starpu_drop_comments(f);
 		int type;
 		ret = fscanf(f, "%d\n", &type);
 		STARPU_ASSERT_MSG(ret == 1, "Incorrect performance model file %s", path);
-		dev_type = _get_enum_type(type);
 		int dev_id;
 		_starpu_drop_comments(f);
 		ret = fscanf(f, "%d\n", &dev_id);
@@ -516,7 +676,7 @@ static void parse_comb(FILE *f, const char *path, struct starpu_perfmodel *model
 		_starpu_drop_comments(f);
 		ret = fscanf(f, "%d\n", &ncores);
 		STARPU_ASSERT_MSG(ret == 1, "Incorrect performance model file %s", path);
-		devices[dev].type = dev_type;
+		devices[dev].type = type;
 		devices[dev].devid = dev_id;
 		devices[dev].ncores = ncores;
 	}
@@ -573,6 +733,43 @@ static int parse_model_file(FILE *f, const char *path, struct starpu_perfmodel *
 }
 
 #ifndef STARPU_SIMGRID
+static void check_per_arch_model(struct starpu_perfmodel *model, int comb, unsigned impl)
+{
+	struct starpu_perfmodel_per_arch *per_arch_model;
+
+	per_arch_model = &model->state->per_arch[comb][impl];
+	/* count the number of elements in the lists */
+	struct starpu_perfmodel_history_list *ptr = NULL;
+	unsigned nentries = 0;
+
+	if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED  || model->type == STARPU_REGRESSION_BASED)
+	{
+		/* Dump the list of all entries in the history */
+		ptr = per_arch_model->list;
+		while(ptr)
+		{
+			nentries++;
+			ptr = ptr->next;
+		}
+	}
+
+	/* header */
+	char archname[STR_SHORT_LENGTH];
+	starpu_perfmodel_get_arch_name(arch_combs[comb], archname,  sizeof(archname), impl);
+	STARPU_ASSERT(strlen(archname)>0);
+	check_reg_model(model, comb, impl);
+
+	/* Dump the history into the model file in case it is necessary */
+	if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED || model->type == STARPU_REGRESSION_BASED)
+	{
+		ptr = per_arch_model->list;
+		while (ptr)
+		{
+			check_history_entry(ptr->entry);
+			ptr = ptr->next;
+		}
+	}
+}
 static void dump_per_arch_model_file(FILE *f, struct starpu_perfmodel *model, int comb, unsigned impl)
 {
 	struct starpu_perfmodel_per_arch *per_arch_model;
@@ -582,7 +779,7 @@ static void dump_per_arch_model_file(FILE *f, struct starpu_perfmodel *model, in
 	struct starpu_perfmodel_history_list *ptr = NULL;
 	unsigned nentries = 0;
 
-	if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED)
+       if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED || model->type == STARPU_REGRESSION_BASED)
 	{
 		/* Dump the list of all entries in the history */
 		ptr = per_arch_model->list;
@@ -603,7 +800,7 @@ static void dump_per_arch_model_file(FILE *f, struct starpu_perfmodel *model, in
 	dump_reg_model(f, model, comb, impl);
 
 	/* Dump the history into the model file in case it is necessary */
-	if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED)
+       if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED || model->type == STARPU_REGRESSION_BASED)
 	{
 		fprintf(f, "# hash\t\tsize\t\tflops\t\tmean (us)\tdev (us)\tsum\t\tsum2\t\tn\n");
 		ptr = per_arch_model->list;
@@ -617,6 +814,39 @@ static void dump_per_arch_model_file(FILE *f, struct starpu_perfmodel *model, in
 	fprintf(f, "\n");
 }
 
+static void check_model(struct starpu_perfmodel *model)
+{
+	int ncombs = model->state->ncombs;
+	STARPU_ASSERT(ncombs >= 0);
+
+	int i, impl, dev;
+	for(i = 0; i < ncombs; i++)
+	{
+		int comb = model->state->combs[i];
+		STARPU_ASSERT(comb >= 0);
+
+		int ndevices = arch_combs[comb]->ndevices;
+		STARPU_ASSERT(ndevices >= 1);
+
+		for(dev = 0; dev < ndevices; dev++)
+		{
+			STARPU_ASSERT(arch_combs[comb]->devices[dev].type >= 0);
+			STARPU_ASSERT(arch_combs[comb]->devices[dev].type <= 5);
+
+			STARPU_ASSERT(arch_combs[comb]->devices[dev].devid >= 0);
+
+			STARPU_ASSERT(arch_combs[comb]->devices[dev].ncores >= 0);
+		}
+
+		int nimpls = model->state->nimpls[comb];
+		STARPU_ASSERT(nimpls >= 1);
+		for (impl = 0; impl < nimpls; impl++)
+		{
+			check_per_arch_model(model, comb, impl);
+		}
+	}
+}
+
 static void dump_model_file(FILE *f, struct starpu_perfmodel *model)
 {
 	fprintf(f, "##################\n");
@@ -644,7 +874,8 @@ static void dump_model_file(FILE *f, struct starpu_perfmodel *model)
 		{
 			fprintf(f, "####################\n");
 			fprintf(f, "# DEV_%d\n", dev);
-			fprintf(f, "# device type (CPU - 0, CUDA - 1, OPENCL - 2, MIC - 3, SCC - 4)\n");
+			fprintf(f, "# device type (CPU - %d, CUDA - %d, OPENCL - %d, MIC - %d, MPI_MS - %d)\n",
+				STARPU_CPU_WORKER, STARPU_CUDA_WORKER, STARPU_OPENCL_WORKER, STARPU_MIC_WORKER, STARPU_MPI_MS_WORKER);
 			fprintf(f, "%u\n", arch_combs[comb]->devices[dev].type);
 
 			fprintf(f, "####################\n");
@@ -670,6 +901,164 @@ static void dump_model_file(FILE *f, struct starpu_perfmodel *model)
 }
 #endif
 
+static void dump_history_entry_xml(FILE *f, struct starpu_perfmodel_history_entry *entry)
+{
+	fprintf(f, "      <entry footprint=\"%08x\" size=\"%lu\" flops=\"%e\" mean=\"%e\" deviation=\"%e\" sum=\"%e\" sum2=\"%e\" nsample=\"%u\"/>\n", entry->footprint, (unsigned long) entry->size, entry->flops, entry->mean, entry->deviation, entry->sum, entry->sum2, entry->nsample);
+}
+
+static void dump_reg_model_xml(FILE *f, struct starpu_perfmodel *model, int comb, int impl)
+{
+	struct starpu_perfmodel_per_arch *per_arch_model;
+
+	per_arch_model = &model->state->per_arch[comb][impl];
+	struct starpu_perfmodel_regression_model *reg_model = &per_arch_model->regression;
+
+	/*
+	 * Linear Regression model
+	 */
+
+	if (model->type == STARPU_REGRESSION_BASED)
+	{
+		fprintf(f, "      <!-- time = alpha size ^ beta -->\n");
+		fprintf(f, "      <l_regression sumlnx=\"%e\" sumlnx2=\"%e\" sumlny=\"%e\" sumlnxlny=\"%e\"", reg_model->sumlnx, reg_model->sumlnx2, reg_model->sumlny, reg_model->sumlnxlny);
+		fprintf(f, " alpha=\"");
+		_starpu_write_double(f, "%e", reg_model->alpha);
+		fprintf(f, "\" beta=\"");
+		_starpu_write_double(f, "%e", reg_model->beta);
+		fprintf(f, "\" nsample=\"%u\" minx=\"%lu\" maxx=\"%lu\"/>\n", reg_model->nsample, reg_model->minx, reg_model->maxx);
+	}
+
+	/*
+	 * Non-Linear Regression model
+	 */
+
+	else if (model->type == STARPU_NL_REGRESSION_BASED)
+	{
+		fprintf(f, "      <!-- time = a size ^b + c -->\n");
+		fprintf(f, "      <nl_regression a=\"");
+		_starpu_write_double(f, "%e", reg_model->a);
+		fprintf(f, "\" b=\"");
+		_starpu_write_double(f, "%e", reg_model->b);
+		fprintf(f, "\" c=\"");
+		_starpu_write_double(f, "%e", reg_model->c);
+		fprintf(f, "\"/>\n");
+	}
+
+	else if (model->type == STARPU_MULTIPLE_REGRESSION_BASED)
+	{
+		if (reg_model->ncoeff==0 || model->ncombinations==0 || model->combinations==NULL)
+			fprintf(f, "      <ml_regression constant=\"nan\"/>\n");
+		else
+		{
+			unsigned i;
+			fprintf(f, "      <ml_regression constant=\"%e\">\n", reg_model->coeff[0]);
+			for (i=0; i < model->ncombinations; i++)
+			{
+				fprintf(f, "        <monomial name=\"");
+				if (model->parameters_names == NULL)
+					fprintf(f, "c%u", i+1);
+				else
+				{
+					unsigned j;
+					int first=1;
+					for(j=0; j < model->nparameters; j++)
+					{
+						if (model->combinations[i][j] > 0)
+						{
+							if (first)
+								first=0;
+							else
+								fprintf(f, "*");
+
+							if(model->parameters_names[j] != NULL)
+								fprintf(f, "%s", model->parameters_names[j]);
+							else
+								fprintf(f, "P%u", j);
+
+							if (model->combinations[i][j] > 1)
+								fprintf(f, "^%d", model->combinations[i][j]);
+						}
+					}
+				}
+				fprintf(f, "\" coef=\"%e\"/>\n", reg_model->coeff[i+1]);
+			}
+			fprintf(f, "      </ml_regression>\n");
+		}
+	}
+}
+
+static void dump_per_arch_model_xml(FILE *f, struct starpu_perfmodel *model, int comb, unsigned impl)
+{
+	struct starpu_perfmodel_per_arch *per_arch_model;
+
+	per_arch_model = &model->state->per_arch[comb][impl];
+	/* count the number of elements in the lists */
+	struct starpu_perfmodel_history_list *ptr = NULL;
+
+	dump_reg_model_xml(f, model, comb, impl);
+
+	/* Dump the history into the model file in case it is necessary */
+	ptr = per_arch_model->list;
+	while (ptr)
+	{
+		dump_history_entry_xml(f, ptr->entry);
+		ptr = ptr->next;
+	}
+}
+
+void starpu_perfmodel_dump_xml(FILE *f, struct starpu_perfmodel *model)
+{
+	fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
+	fprintf(f, "<!DOCTYPE StarPUPerfmodel SYSTEM \"starpu-perfmodel.dtd\">\n");
+	fprintf(f, "<!-- symbol %s -->\n", model->symbol);
+	fprintf(f, "<!-- All times in us -->\n");
+	fprintf(f, "<perfmodel version=\"%u\">\n", _STARPU_PERFMODEL_VERSION);
+
+	STARPU_PTHREAD_RWLOCK_RDLOCK(&model->state->model_rwlock);
+	int ncombs = model->state->ncombs;
+	int i, impl, dev;
+
+	for(i = 0; i < ncombs; i++)
+	{
+		int comb = model->state->combs[i];
+		int ndevices = arch_combs[comb]->ndevices;
+
+		fprintf(f, "  <combination>\n");
+		for(dev = 0; dev < ndevices; dev++)
+		{
+			const char *type;
+			switch (arch_combs[comb]->devices[dev].type) {
+				case STARPU_CPU_WORKER: type = "CPU"; break;
+				case STARPU_CUDA_WORKER: type = "CUDA"; break;
+				case STARPU_OPENCL_WORKER: type = "OpenCL"; break;
+				case STARPU_MIC_WORKER: type = "MIC"; break;
+				case STARPU_MPI_MS_WORKER: type = "MPI_MS"; break;
+				default: STARPU_ASSERT(0);
+			}
+			fprintf(f, "    <device type=\"%s\" id=\"%d\"",
+					type,
+					arch_combs[comb]->devices[dev].devid);
+			if (arch_combs[comb]->devices[dev].type == STARPU_CPU_WORKER)
+				fprintf(f, " ncores=\"%d\"",
+						arch_combs[comb]->devices[dev].ncores);
+			fprintf(f, "/>\n");
+		}
+		int nimpls = model->state->nimpls[comb];
+		for (impl = 0; impl < nimpls; impl++)
+		{
+			fprintf(f, "    <implementation id=\"%d\">\n", impl);
+			char archname[STR_SHORT_LENGTH];
+			starpu_perfmodel_get_arch_name(arch_combs[comb], archname,  sizeof(archname), impl);
+			fprintf(f, "      <!-- %s -->\n", archname);
+			dump_per_arch_model_xml(f, model, comb, impl);
+			fprintf(f, "    </implementation>\n");
+		}
+		fprintf(f, "  </combination>\n");
+	}
+	STARPU_PTHREAD_RWLOCK_UNLOCK(&model->state->model_rwlock);
+	fprintf(f, "</perfmodel>\n");
+}
+
 void _starpu_perfmodel_realloc(struct starpu_perfmodel *model, int nb)
 {
 	int i;
@@ -696,7 +1085,7 @@ void _starpu_perfmodel_realloc(struct starpu_perfmodel *model, int nb)
 void starpu_perfmodel_init(struct starpu_perfmodel *model)
 {
 	int already_init;
-	int i, ncombs;
+	int ncombs;
 
 	STARPU_ASSERT(model);
 
@@ -724,21 +1113,13 @@ void starpu_perfmodel_init(struct starpu_perfmodel *model)
 	STARPU_PTHREAD_RWLOCK_RDLOCK(&arch_combs_mutex);
 	model->state->ncombs_set = ncombs = nb_arch_combs;
 	STARPU_PTHREAD_RWLOCK_UNLOCK(&arch_combs_mutex);
-	_STARPU_MALLOC(model->state->per_arch, ncombs*sizeof(struct starpu_perfmodel_per_arch*));
-	_STARPU_MALLOC(model->state->per_arch_is_set, ncombs*sizeof(int*));
-	_STARPU_MALLOC(model->state->nimpls, ncombs*sizeof(int));
-	_STARPU_MALLOC(model->state->nimpls_set, ncombs*sizeof(int));
+	_STARPU_CALLOC(model->state->per_arch, ncombs, sizeof(struct starpu_perfmodel_per_arch*));
+	_STARPU_CALLOC(model->state->per_arch_is_set, ncombs, sizeof(int*));
+	_STARPU_CALLOC(model->state->nimpls, ncombs, sizeof(int));
+	_STARPU_CALLOC(model->state->nimpls_set, ncombs, sizeof(int));
 	_STARPU_MALLOC(model->state->combs, ncombs*sizeof(int));
 	model->state->ncombs = 0;
 
-	for(i = 0; i < ncombs; i++)
-	{
-		model->state->per_arch[i] = NULL;
-		model->state->per_arch_is_set[i] = NULL;
-		model->state->nimpls[i] = 0;
-		model->state->nimpls_set[i] = 0;
-	}
-
 	/* add the model to a linked list */
 	struct _starpu_perfmodel *node = _starpu_perfmodel_new();
 
@@ -792,6 +1173,7 @@ static void save_history_based_model(struct starpu_perfmodel *model)
 	STARPU_ASSERT_MSG(f, "Could not save performance model %s\n", path);
 
 	locked = _starpu_fwrlock(f) == 0;
+	check_model(model);
 	_starpu_fftruncate(f, 0);
 	dump_model_file(f, model);
 	if (locked)
@@ -845,11 +1227,13 @@ void _starpu_initialize_registered_performance_models(void)
 	unsigned i;
 	for(i = 0; i < conf->topology.nhwmicdevices; i++)
 		nmic += conf->topology.nhwmiccores[i];
-	unsigned nscc = conf->topology.nhwscc;
+	unsigned nmpi = 0;
+	for(i = 0; i < conf->topology.nhwmpidevices; i++)
+		nmpi += conf->topology.nhwmpicores[i];
 
-	// We used to allocate 2**(ncores + ncuda + nopencl + nmic + nscc), this is too big
-	// We now allocate only 2*(ncores + ncuda + nopencl + nmic + nscc), and reallocate when necessary in starpu_perfmodel_arch_comb_add
-	nb_arch_combs = 2 * (ncores + ncuda + nopencl + nmic + nscc);
+	// We used to allocate 2**(ncores + ncuda + nopencl + nmic + nmpi), this is too big
+	// We now allocate only 2*(ncores + ncuda + nopencl + nmic + nmpi), and reallocate when necessary in starpu_perfmodel_arch_comb_add
+	nb_arch_combs = 2 * (ncores + ncuda + nopencl + nmic + nmpi);
 	_STARPU_MALLOC(arch_combs, nb_arch_combs*sizeof(struct starpu_perfmodel_arch*));
 	current_arch_comb = 0;
 	historymaxerror = starpu_get_env_number_default("STARPU_HISTORY_MAX_ERROR", STARPU_HISTORYMAXERROR);
@@ -858,7 +1242,7 @@ void _starpu_initialize_registered_performance_models(void)
 	ignore_devid[STARPU_CUDA_WORKER] = starpu_get_env_number_default("STARPU_PERF_MODEL_HOMOGENEOUS_CUDA", 0);
 	ignore_devid[STARPU_OPENCL_WORKER] = starpu_get_env_number_default("STARPU_PERF_MODEL_HOMOGENEOUS_OPENCL", 0);
 	ignore_devid[STARPU_MIC_WORKER] = starpu_get_env_number_default("STARPU_PERF_MODEL_HOMOGENEOUS_MIC", 0);
-	ignore_devid[STARPU_SCC_WORKER] = starpu_get_env_number_default("STARPU_PERF_MODEL_HOMOGENEOUS_SCC", 0);
+	ignore_devid[STARPU_MPI_MS_WORKER] = starpu_get_env_number_default("STARPU_PERF_MODEL_HOMOGENEOUS_MPI_MS", 0);
 }
 
 void _starpu_deinitialize_performance_model(struct starpu_perfmodel *model)
@@ -876,8 +1260,8 @@ void _starpu_deinitialize_performance_model(struct starpu_perfmodel *model)
 					struct starpu_perfmodel_per_arch *archmodel = &model->state->per_arch[i][impl];
 					if (archmodel->history)
 					{
-						struct starpu_perfmodel_history_list *list, *plist;
-						struct starpu_perfmodel_history_table *entry, *tmp;
+						struct starpu_perfmodel_history_list *list;
+						struct starpu_perfmodel_history_table *entry=NULL, *tmp=NULL;
 
 						HASH_ITER(hh, archmodel->history, entry, tmp)
 						{
@@ -889,6 +1273,7 @@ void _starpu_deinitialize_performance_model(struct starpu_perfmodel *model)
 						list = archmodel->list;
 						while (list)
 						{
+							struct starpu_perfmodel_history_list *plist;
 							free(list->entry);
 							plist = list;
 							list = list->next;
@@ -970,11 +1355,11 @@ void _starpu_load_history_based_model(struct starpu_perfmodel *model, unsigned s
 	if(!model->is_loaded)
 	{
 		char path[STR_LONG_LENGTH];
-		starpu_perfmodel_get_model_path(model->symbol, path, sizeof(path));
-
 		// Check if a symbol is defined before trying to load the model from a file
 		STARPU_ASSERT_MSG(model->symbol, "history-based performance models must have a symbol");
 
+		starpu_perfmodel_get_model_path(model->symbol, path, sizeof(path));
+
 		_STARPU_DEBUG("Opening performance model file %s for model %s ...\n", path, model->symbol);
 
 		unsigned calibrate_flag = _starpu_get_calibrate_flag();
@@ -1023,27 +1408,29 @@ void starpu_perfmodel_directory(FILE *output)
  * the performance model files */
 int starpu_perfmodel_list(FILE *output)
 {
-#if !defined(_WIN32) || defined(__MINGW32__) || defined(__CYGWIN__)
+#ifdef HAVE_SCANDIR
         char *path;
-        DIR *dp;
-        struct dirent *ep;
+	struct dirent **list;
+	int n;
 
 	path = _starpu_get_perf_model_dir_codelet();
-        dp = opendir(path);
-        if (dp != NULL)
-	{
-                while ((ep = readdir(dp)))
-		{
-                        if (strcmp(ep->d_name, ".") && strcmp(ep->d_name, ".."))
-                                fprintf(output, "file: <%s>\n", ep->d_name);
-                }
-                closedir (dp);
-        }
-        else
+	n = scandir(path, &list, NULL, alphasort);
+	if (n < 0)
 	{
 		_STARPU_DISP("Could not open the perfmodel directory <%s>: %s\n", path, strerror(errno));
-        }
-	return 0;
+		return 1;
+	}
+	else
+	{
+		int i;
+		for (i = 0; i < n; i++) {
+			if (strcmp(list[i]->d_name, ".") && strcmp(list[i]->d_name, ".."))
+				fprintf(output, "file: <%s>\n", list[i]->d_name);
+			free(list[i]);
+		}
+		free(list);
+		return 0;
+	}
 #else
 	_STARPU_MSG("Listing perfmodels is not implemented on pure Windows yet\n");
 	return 1;
@@ -1153,8 +1540,8 @@ char* starpu_perfmodel_get_archtype_name(enum starpu_worker_archtype archtype)
 		case(STARPU_MIC_WORKER):
 			return "mic";
 			break;
-		case(STARPU_SCC_WORKER):
-			return "scc";
+		case(STARPU_MPI_MS_WORKER):
+			return "mpi_ms";
 			break;
 		default:
 			STARPU_ABORT();
@@ -1170,7 +1557,7 @@ void starpu_perfmodel_get_arch_name(struct starpu_perfmodel_arch* arch, char *ar
 	STARPU_ASSERT(comb != -1);
 	char devices[STR_VERY_LONG_LENGTH];
 	int written = 0;
-	strcpy(devices, "");
+	devices[0] = '\0';
 	for(i=0 ; i<arch->ndevices ; i++)
 	{
 		written += snprintf(devices + written, sizeof(devices)-written, "%s%d%s", starpu_perfmodel_get_archtype_name(arch->devices[i].type), arch->devices[i].devid, i != arch->ndevices-1 ? "_":"");
@@ -1281,6 +1668,58 @@ docal:
 	return exp;
 }
 
+double _starpu_multiple_regression_based_job_expected_perf(struct starpu_perfmodel *model, struct starpu_perfmodel_arch* arch, struct _starpu_job *j, unsigned nimpl)
+{
+	int comb;
+	double expected_duration=NAN;
+
+	struct starpu_perfmodel_regression_model *reg_model = NULL;
+	comb = starpu_perfmodel_arch_comb_get(arch->ndevices, arch->devices);
+	if(comb == -1)
+		goto docal;
+	if (model->state->per_arch[comb] == NULL)
+		// The model has not been executed on this combination
+		goto docal;
+	reg_model = &model->state->per_arch[comb][nimpl].regression;
+	if (reg_model->coeff == NULL)
+		goto docal;
+
+	double *parameters;
+	_STARPU_MALLOC(parameters, model->nparameters*sizeof(double));
+	model->parameters(j->task, parameters);
+	expected_duration=reg_model->coeff[0];
+	unsigned i;
+	for (i=0; i < model->ncombinations; i++)
+	{
+		double parameter_value=1.;
+		unsigned k;
+		for (k=0; k < model->nparameters; k++)
+			parameter_value *= pow(parameters[k],model->combinations[i][k]);
+
+		expected_duration += reg_model->coeff[i+1]*parameter_value;
+	}
+
+docal:
+	STARPU_HG_DISABLE_CHECKING(model->benchmarking);
+	if (isnan(expected_duration) && !model->benchmarking)
+	{
+		char archname[STR_SHORT_LENGTH];
+
+		starpu_perfmodel_get_arch_name(arch, archname, sizeof(archname), nimpl);
+		_STARPU_DISP("Warning: model %s is not calibrated enough for %s, forcing calibration for this run. Use the STARPU_CALIBRATE environment variable to control this.\n", model->symbol, archname);
+		_starpu_set_calibrate_flag(1);
+		model->benchmarking = 1;
+	}
+
+	// In the unlikely event that predicted duration is negative
+	// in case multiple linear regression is not so accurate
+	if (expected_duration < 0 )
+		expected_duration = 0.00001;
+
+	//Make sure that the injected time is in milliseconds
+	return expected_duration;
+}
+
 double _starpu_history_based_job_expected_perf(struct starpu_perfmodel *model, struct starpu_perfmodel_arch* arch, struct _starpu_job *j,unsigned nimpl)
 {
 	int comb;
@@ -1291,6 +1730,7 @@ double _starpu_history_based_job_expected_perf(struct starpu_perfmodel *model, s
 	uint32_t key;
 
 	comb = starpu_perfmodel_arch_comb_get(arch->ndevices, arch->devices);
+	key = _starpu_compute_buffers_footprint(model, arch, nimpl, j);
 	if(comb == -1)
 		goto docal;
 	if (model->state->per_arch[comb] == NULL)
@@ -1299,35 +1739,62 @@ double _starpu_history_based_job_expected_perf(struct starpu_perfmodel *model, s
 
 	per_arch_model = &model->state->per_arch[comb][nimpl];
 
-	key = _starpu_compute_buffers_footprint(model, arch, nimpl, j);
 	STARPU_PTHREAD_RWLOCK_RDLOCK(&model->state->model_rwlock);
 	history = per_arch_model->history;
 	HASH_FIND_UINT32_T(history, &key, elt);
 	entry = (elt == NULL) ? NULL : elt->history_entry;
+	STARPU_ASSERT_MSG(!entry || entry->mean >= 0, "entry=%p, entry->mean=%lf\n", entry, entry?entry->mean:NAN);
 	STARPU_PTHREAD_RWLOCK_UNLOCK(&model->state->model_rwlock);
 
 	/* Here helgrind would shout that this is unprotected access.
 	 * We do not care about racing access to the mean, we only want
 	 * a good-enough estimation */
 
-	if (entry && entry->nsample >= _starpu_calibration_minimum)
-		/* TODO: report differently if we've scheduled really enough
-		 * of that task and the scheduler should perhaps put it aside */
-		/* Calibrated enough */
-		exp = entry->mean;
+	if (entry && entry->nsample)
+	{
+#ifdef STARPU_SIMGRID
+		if (entry->nsample < _starpu_calibration_minimum)
+		{
+			char archname[STR_SHORT_LENGTH];
+			starpu_perfmodel_get_arch_name(arch, archname, sizeof(archname), nimpl);
+
+			_STARPU_DISP("Warning: model %s is not calibrated enough for %s size %ld footprint %x (only %u measurements). Using it anyway for the simulation\n", model->symbol, archname, j->task?(long int)_starpu_job_get_data_size(model, arch, nimpl, j):-1, key, entry->nsample);
+		}
+#else
+		if (entry->nsample >= _starpu_calibration_minimum)
+#endif
+		{
+			STARPU_ASSERT_MSG(entry->mean >= 0, "entry->mean=%lf\n", entry->mean);
+			/* TODO: report differently if we've scheduled really enough
+			 * of that task and the scheduler should perhaps put it aside */
+			/* Calibrated enough */
+			exp = entry->mean;
+		}
+	}
 
 docal:
+#ifdef STARPU_SIMGRID
+	if (isnan(exp)) {
+		char archname[STR_SHORT_LENGTH];
+		starpu_perfmodel_get_arch_name(arch, archname, sizeof(archname), nimpl);
+
+		_STARPU_DISP("Warning: model %s is not calibrated at all for %s size %ld footprint %x. Assuming it can not work there\n", model->symbol, archname, j->task?(long int)_starpu_job_get_data_size(model, arch, nimpl, j):-1, key);
+		exp = 0.;
+	}
+#else
 	STARPU_HG_DISABLE_CHECKING(model->benchmarking);
 	if (isnan(exp) && !model->benchmarking)
 	{
 		char archname[STR_SHORT_LENGTH];
 
 		starpu_perfmodel_get_arch_name(arch, archname, sizeof(archname), nimpl);
-		_STARPU_DISP("Warning: model %s is not calibrated enough for %s size %ld (only %u measurements), forcing calibration for this run. Use the STARPU_CALIBRATE environment variable to control this.\n", model->symbol, archname, j->task?(long int)_starpu_job_get_data_size(model, arch, nimpl, j):-1, entry ? entry->nsample : 0);
+		_STARPU_DISP("Warning: model %s is not calibrated enough for %s size %ld footprint %x (only %u measurements), forcing calibration for this run. Use the STARPU_CALIBRATE environment variable to control this.\n", model->symbol, archname, j->task?(long int)_starpu_job_get_data_size(model, arch, nimpl, j):-1, key, entry ? entry->nsample : 0);
 		_starpu_set_calibrate_flag(1);
 		model->benchmarking = 1;
 	}
+#endif
 
+	STARPU_ASSERT_MSG(isnan(exp)||exp >= 0, "exp=%lf\n", exp);
 	return exp;
 }
 
@@ -1351,6 +1818,7 @@ int _starpu_perfmodel_create_comb_if_needed(struct starpu_perfmodel_arch* arch)
 
 void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfmodel *model, struct starpu_perfmodel_arch* arch, unsigned cpuid STARPU_ATTRIBUTE_UNUSED, double measured, unsigned impl)
 {
+	STARPU_ASSERT_MSG(measured >= 0, "measured=%lf\n", measured);
 	if (model)
 	{
 		int c;
@@ -1393,7 +1861,7 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 			model->state->per_arch_is_set[comb][impl] = 1;
 		}
 
-		if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED)
+		if (model->type == STARPU_HISTORY_BASED || model->type == STARPU_NL_REGRESSION_BASED || model->type == STARPU_REGRESSION_BASED)
 		{
 			struct starpu_perfmodel_history_entry *entry;
 			struct starpu_perfmodel_history_table *elt;
@@ -1408,7 +1876,7 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 			if (!entry)
 			{
 				/* this is the first entry with such a footprint */
-				_STARPU_MALLOC(entry, sizeof(struct starpu_perfmodel_history_entry));
+				_STARPU_CALLOC(entry, 1, sizeof(struct starpu_perfmodel_history_entry));
 
 				/* Tell  helgrind that we do not care about
 				 * racing access to the sampling, we only want a
@@ -1467,7 +1935,7 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 
 					unsigned n = entry->nsample;
 					entry->mean = entry->sum / n;
-					entry->deviation = sqrt((entry->sum2 - (entry->sum*entry->sum)/n)/n);
+					entry->deviation = sqrt((fabs(entry->sum2 - (entry->sum*entry->sum)/n))/n);
 				}
 
 				if (j->task->flops != 0.)
@@ -1517,6 +1985,26 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 			}
 		}
 
+		if (model->type == STARPU_MULTIPLE_REGRESSION_BASED)
+		{
+			struct starpu_perfmodel_history_entry *entry;
+			struct starpu_perfmodel_history_list **list;
+			list = &per_arch_model->list;
+
+			_STARPU_CALLOC(entry, 1, sizeof(struct starpu_perfmodel_history_entry));
+			_STARPU_MALLOC(entry->parameters, model->nparameters*sizeof(double));
+			model->parameters(j->task, entry->parameters);
+			entry->tag = j->task->tag_id;
+			STARPU_ASSERT(measured >= 0);
+			entry->duration = measured;
+
+			struct starpu_perfmodel_history_list *link;
+			_STARPU_MALLOC(link, sizeof(struct starpu_perfmodel_history_list));
+			link->next = *list;
+			link->entry = entry;
+			*list = link;
+		}
+
 #ifdef STARPU_MODEL_DEBUG
 		struct starpu_task *task = j->task;
 		starpu_perfmodel_debugfilepath(model, arch_combs[comb], per_arch_model->debug_path, STR_LONG_LENGTH, impl);
@@ -1525,6 +2013,7 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 		if (f == NULL)
 		{
 			_STARPU_DISP("Error <%s> when opening file <%s>\n", strerror(errno), per_arch_model->debug_path);
+			STARPU_PTHREAD_RWLOCK_UNLOCK(&model->state->model_rwlock);
 			return;
 		}
 		locked = _starpu_fwrlock(f) == 0;
