@@ -1,10 +1,11 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2011-2016                                Inria
+ * Copyright (C) 2011-2017                                Inria
  * Copyright (C) 2013                                     Simon Archipoff
- * Copyright (C) 2008-2018                                Université de Bordeaux
- * Copyright (C) 2010-2015,2017                           CNRS
+ * Copyright (C) 2008-2019                                Université de Bordeaux
+ * Copyright (C) 2010-2017, 2019                          CNRS
  * Copyright (C) 2013                                     Thibaut Lambert
+ * Copyright (C) 2016                                     Uppsala University
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -25,6 +26,7 @@
 #include <profiling/profiling.h>
 #include <common/barrier.h>
 #include <core/debug.h>
+#include <core/task.h>
 
 static int use_prefetch = 0;
 static double idle[STARPU_NMAXWORKERS];
@@ -54,16 +56,20 @@ static struct starpu_sched_policy *predefined_policies[] =
 {
 	&_starpu_sched_modular_eager_policy,
 	&_starpu_sched_modular_eager_prefetching_policy,
+	&_starpu_sched_modular_gemm_policy,
 	&_starpu_sched_modular_prio_policy,
 	&_starpu_sched_modular_prio_prefetching_policy,
 	&_starpu_sched_modular_random_policy,
 	&_starpu_sched_modular_random_prio_policy,
 	&_starpu_sched_modular_random_prefetching_policy,
 	&_starpu_sched_modular_random_prio_prefetching_policy,
-	//&_starpu_sched_modular_ws_policy,
+	&_starpu_sched_modular_parallel_random_policy,
+	&_starpu_sched_modular_parallel_random_prio_policy,
+	&_starpu_sched_modular_ws_policy,
 	&_starpu_sched_modular_heft_policy,
 	&_starpu_sched_modular_heft_prio_policy,
 	&_starpu_sched_modular_heft2_policy,
+	&_starpu_sched_modular_parallel_heft_policy,
 	&_starpu_sched_eager_policy,
 	&_starpu_sched_prio_policy,
 	&_starpu_sched_random_policy,
@@ -76,7 +82,11 @@ static struct starpu_sched_policy *predefined_policies[] =
 	&_starpu_sched_dmda_sorted_decision_policy,
 	&_starpu_sched_parallel_heft_policy,
 	&_starpu_sched_peager_policy,
+	&_starpu_sched_heteroprio_policy,
 	&_starpu_sched_graph_test_policy,
+#ifdef STARPU_HAVE_HWLOC
+	//&_starpu_sched_tree_heft_hierarchical_policy,
+#endif
 	NULL
 };
 
@@ -194,7 +204,7 @@ struct starpu_sched_policy *_starpu_select_sched_policy(struct _starpu_machine_c
 		return selected_policy;
 
 	/* If no policy was specified, we use the eager policy by default */
-	return &_starpu_sched_eager_policy;
+	return &_starpu_sched_lws_policy;
 }
 
 void _starpu_init_sched_policy(struct _starpu_machine_config *config, struct _starpu_sched_ctx *sched_ctx, struct starpu_sched_policy *selected_policy)
@@ -212,9 +222,20 @@ void _starpu_init_sched_policy(struct _starpu_machine_config *config, struct _st
 
 	load_sched_policy(selected_policy, sched_ctx);
 
-	_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+	if (starpu_get_env_number_default("STARPU_WORKER_TREE", 0))
+	{
+#ifdef STARPU_HAVE_HWLOC
+		sched_ctx->sched_policy->worker_type = STARPU_WORKER_TREE;
+#else
+		_STARPU_DISP("STARPU_WORKER_TREE ignored, please rebuild StarPU with hwloc support to enable it.");
+#endif
+	}
+	starpu_sched_ctx_create_worker_collection(sched_ctx->id,
+						  sched_ctx->sched_policy->worker_type);
+
+	_STARPU_SCHED_BEGIN;
 	sched_ctx->sched_policy->init_sched(sched_ctx->id);
-	_STARPU_TRACE_WORKER_SCHEDULING_POP;
+	_STARPU_SCHED_END;
 }
 
 void _starpu_deinit_sched_policy(struct _starpu_sched_ctx *sched_ctx)
@@ -222,10 +243,11 @@ void _starpu_deinit_sched_policy(struct _starpu_sched_ctx *sched_ctx)
 	struct starpu_sched_policy *policy = sched_ctx->sched_policy;
 	if (policy->deinit_sched)
 	{
-		_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+		_STARPU_SCHED_BEGIN;
 		policy->deinit_sched(sched_ctx->id);
-		_STARPU_TRACE_WORKER_SCHEDULING_POP;
+		_STARPU_SCHED_END;
 	}
+	starpu_sched_ctx_delete_worker_collection(sched_ctx->id);
 }
 
 void _starpu_sched_task_submit(struct starpu_task *task)
@@ -235,9 +257,9 @@ void _starpu_sched_task_submit(struct starpu_task *task)
 		return;
 	if (!sched_ctx->sched_policy->submit_hook)
 		return;
-	_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+	_STARPU_SCHED_BEGIN;
 	sched_ctx->sched_policy->submit_hook(task);
-	_STARPU_TRACE_WORKER_SCHEDULING_POP;
+	_STARPU_SCHED_END;
 }
 
 void _starpu_sched_do_schedule(unsigned sched_ctx_id)
@@ -247,24 +269,26 @@ void _starpu_sched_do_schedule(unsigned sched_ctx_id)
 		return;
 	if (!sched_ctx->sched_policy->do_schedule)
 		return;
-	_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+	_STARPU_SCHED_BEGIN;
 	sched_ctx->sched_policy->do_schedule(sched_ctx_id);
-	_STARPU_TRACE_WORKER_SCHEDULING_POP;
+	_STARPU_SCHED_END;
 }
 
 static void _starpu_push_task_on_specific_worker_notify_sched(struct starpu_task *task, struct _starpu_worker *worker, int workerid, int perf_workerid)
 {
 	/* if we push a task on a specific worker, notify all the sched_ctxs the worker belongs to */
-	struct _starpu_sched_ctx *sched_ctx;
-	struct _starpu_sched_ctx_list *l = NULL;
-        for (l = worker->sched_ctx_list; l; l = l->next)
-        {
-		sched_ctx = _starpu_get_sched_ctx_struct(l->sched_ctx);
+	struct _starpu_sched_ctx_list_iterator list_it;
+
+	_starpu_sched_ctx_list_iterator_init(worker->sched_ctx_list, &list_it);
+	while (_starpu_sched_ctx_list_iterator_has_next(&list_it))
+	{
+		struct _starpu_sched_ctx_elt *e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+		struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(e->sched_ctx);
 		if (sched_ctx->sched_policy != NULL && sched_ctx->sched_policy->push_task_notify)
 		{
-			_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+			_STARPU_SCHED_BEGIN;
 			sched_ctx->sched_policy->push_task_notify(task, workerid, perf_workerid, sched_ctx->id);
-			_STARPU_TRACE_WORKER_SCHEDULING_POP;
+			_STARPU_SCHED_END;
 		}
 	}
 }
@@ -279,23 +303,20 @@ static int _starpu_push_task_on_specific_worker(struct starpu_task *task, int wo
 	/* Is this a basic worker or a combined worker ? */
 	int is_basic_worker = (workerid < nbasic_workers);
 
-	unsigned memory_node;
 	struct _starpu_worker *worker = NULL;
 	struct _starpu_combined_worker *combined_worker = NULL;
 
 	if (is_basic_worker)
 	{
 		worker = _starpu_get_worker_struct(workerid);
-		memory_node = worker->memory_node;
 	}
 	else
 	{
 		combined_worker = _starpu_get_combined_worker_struct(workerid);
-		memory_node = combined_worker->memory_node;
 	}
 
 	if (use_prefetch)
-		starpu_prefetch_task_input_on_node(task, memory_node);
+		starpu_prefetch_task_input_for(task, workerid);
 
 	if (is_basic_worker)
 		_starpu_push_task_on_specific_worker_notify_sched(task, worker, workerid, workerid);
@@ -316,13 +337,13 @@ static int _starpu_push_task_on_specific_worker(struct starpu_task *task, int wo
 #ifdef STARPU_USE_SC_HYPERVISOR
 	starpu_sched_ctx_call_pushed_task_cb(workerid, task->sched_ctx);
 #endif //STARPU_USE_SC_HYPERVISOR
-	unsigned i;
 	if (is_basic_worker)
 	{
 		unsigned node = starpu_worker_get_memory_node(workerid);
 		if (_starpu_task_uses_multiformat_handles(task))
 		{
 			unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+			unsigned i;
 			for (i = 0; i < nbuffers; i++)
 			{
 				struct starpu_task *conversion_task;
@@ -394,7 +415,11 @@ static int _starpu_push_task_on_specific_worker(struct starpu_task *task, int wo
 int _starpu_push_task(struct _starpu_job *j)
 {
 	if(j->task->prologue_callback_func)
+	{
+		_starpu_set_current_task(j->task);
 		j->task->prologue_callback_func(j->task->prologue_callback_arg);
+		_starpu_set_current_task(NULL);
+	}
 
 	return _starpu_repush_task(j);
 }
@@ -403,7 +428,6 @@ int _starpu_repush_task(struct _starpu_job *j)
 {
 	struct starpu_task *task = j->task;
 	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(task->sched_ctx);
-	unsigned nworkers = 0;
 	int ret;
 
 	_STARPU_LOG_IN();
@@ -413,19 +437,20 @@ int _starpu_repush_task(struct _starpu_job *j)
 	task->status = STARPU_TASK_READY;
 	STARPU_AYU_ADDTOTASKQUEUE(j->job_id, -1);
 	/* if the context does not have any workers save the tasks in a temp list */
-	if ((task->cl != NULL && task->cl->where != STARPU_NOWHERE) && (!sched_ctx->is_initial_sched))
+	if ((task->cl != NULL && task->where != STARPU_NOWHERE) && (!sched_ctx->is_initial_sched))
 	{
 		/*if there are workers in the ctx that are not able to execute tasks
 		  we consider the ctx empty */
-		nworkers = _starpu_nworkers_able_to_execute_task(task, sched_ctx);
+		unsigned nworkers = _starpu_nworkers_able_to_execute_task(task, sched_ctx);
 
 		if(nworkers == 0)
 		{
-			STARPU_PTHREAD_MUTEX_LOCK(&sched_ctx->empty_ctx_mutex);
+			_starpu_sched_ctx_lock_write(sched_ctx->id);
 			starpu_task_list_push_front(&sched_ctx->empty_ctx_tasks, task);
-			STARPU_PTHREAD_MUTEX_UNLOCK(&sched_ctx->empty_ctx_mutex);
+			_starpu_sched_ctx_unlock_write(sched_ctx->id);
 #ifdef STARPU_USE_SC_HYPERVISOR
-			if(sched_ctx->id != 0 && sched_ctx->perf_counters != NULL && sched_ctx->perf_counters->notify_empty_ctx)
+			if(sched_ctx->id != 0 && sched_ctx->perf_counters != NULL
+			   && sched_ctx->perf_counters->notify_empty_ctx)
 			{
 				_STARPU_TRACE_HYPERVISOR_BEGIN();
 				sched_ctx->perf_counters->notify_empty_ctx(sched_ctx->id, task);
@@ -442,11 +467,15 @@ int _starpu_repush_task(struct _starpu_job *j)
 	/* in case there is no codelet associated to the task (that's a control
 	 * task), we directly execute its callback and enforce the
 	 * corresponding dependencies */
-	if (task->cl == NULL || task->cl->where == STARPU_NOWHERE)
+	if (task->cl == NULL || task->where == STARPU_NOWHERE)
 	{
 		task->status = STARPU_TASK_RUNNING;
 		if (task->prologue_callback_pop_func)
+		{
+			_starpu_set_current_task(task);
 			task->prologue_callback_pop_func(task->prologue_callback_pop_arg);
+			_starpu_set_current_task(NULL);
+		}
 
 		if (task->cl && task->cl->specific_nodes)
 		{
@@ -487,11 +516,12 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 
 		if (nworkers == 0)
 		{
-			STARPU_PTHREAD_MUTEX_LOCK(&sched_ctx->empty_ctx_mutex);
+			_starpu_sched_ctx_lock_write(sched_ctx->id);
 			starpu_task_list_push_back(&sched_ctx->empty_ctx_tasks, task);
-			STARPU_PTHREAD_MUTEX_UNLOCK(&sched_ctx->empty_ctx_mutex);
+			_starpu_sched_ctx_unlock_write(sched_ctx->id);
 #ifdef STARPU_USE_SC_HYPERVISOR
-			if(sched_ctx->id != 0 && sched_ctx->perf_counters != NULL && sched_ctx->perf_counters->notify_empty_ctx)
+			if(sched_ctx->id != 0 && sched_ctx->perf_counters != NULL
+			   && sched_ctx->perf_counters->notify_empty_ctx)
 			{
 				_STARPU_TRACE_HYPERVISOR_BEGIN();
 				sched_ctx->perf_counters->notify_empty_ctx(sched_ctx->id, task);
@@ -508,9 +538,8 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 	int ret = 0;
 	if (STARPU_UNLIKELY(task->execute_on_a_specific_worker))
 	{
-		unsigned node = starpu_worker_get_memory_node(task->workerid);
 		if (starpu_get_prefetch_flag())
-			starpu_prefetch_task_input_on_node(task, node);
+			starpu_prefetch_task_input_for(task, task->workerid);
 
 		ret = _starpu_push_task_on_specific_worker(task, task->workerid);
 	}
@@ -523,16 +552,14 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 		 * prefetch before the scheduling decision. */
 		if (starpu_get_prefetch_flag())
 		{
-			if (task->cl->where == STARPU_CPU && config->cpus_nodeid >= 0)
+			if (task->where == STARPU_CPU && config->cpus_nodeid >= 0)
 				starpu_prefetch_task_input_on_node(task, config->cpus_nodeid);
-			else if (task->cl->where == STARPU_CUDA && config->cuda_nodeid >= 0)
+			else if (task->where == STARPU_CUDA && config->cuda_nodeid >= 0)
 				starpu_prefetch_task_input_on_node(task, config->cuda_nodeid);
-			else if (task->cl->where == STARPU_OPENCL && config->opencl_nodeid >= 0)
+			else if (task->where == STARPU_OPENCL && config->opencl_nodeid >= 0)
 				starpu_prefetch_task_input_on_node(task, config->opencl_nodeid);
-			else if (task->cl->where == STARPU_MIC && config->mic_nodeid >= 0)
+			else if (task->where == STARPU_MIC && config->mic_nodeid >= 0)
 				starpu_prefetch_task_input_on_node(task, config->mic_nodeid);
-			else if (task->cl->where == STARPU_SCC && config->scc_nodeid >= 0)
-				starpu_prefetch_task_input_on_node(task, config->scc_nodeid);
 		}
 
 		if(!sched_ctx->sched_policy)
@@ -545,7 +572,7 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 			else
 			{
 				struct starpu_worker_collection *workers = sched_ctx->workers;
-				
+
 				struct _starpu_job *job = _starpu_get_job_associated_to_task(task);
 				job->task_size = workers->nworkers;
 				job->combined_workerid = -1; // workerid; its a ctx not combined worker
@@ -555,17 +582,22 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 				STARPU_PTHREAD_BARRIER_INIT(&job->after_work_barrier, NULL, workers->nworkers);
 				job->after_work_busy_barrier = workers->nworkers;
 
-				unsigned workerid;
 				struct starpu_sched_ctx_iterator it;
 				if(workers->init_iterator)
 					workers->init_iterator(workers, &it);
 
 				while(workers->has_next(workers, &it))
 				{
-					workerid = workers->get_next(workers, &it);
-					struct starpu_task *alias = starpu_task_dup(task);
-					_STARPU_TRACE_JOB_PUSH(alias, alias->priority > 0);
-					alias->destroy = 1;
+					unsigned workerid = workers->get_next(workers, &it);
+					struct starpu_task *alias;
+					if (job->task_size > 1)
+					{
+						alias = starpu_task_dup(task);
+						_STARPU_TRACE_JOB_PUSH(alias, alias->priority > 0);
+						alias->destroy = 1;
+					}
+					else
+						alias = task;
 					ret |= _starpu_push_task_on_specific_worker(alias, workerid);
 				}
 			}
@@ -574,19 +606,29 @@ int _starpu_push_task_to_workers(struct starpu_task *task)
 		{
 			STARPU_ASSERT(sched_ctx->sched_policy->push_task);
 			/* check out if there are any workers in the context */
-			starpu_pthread_rwlock_t *changing_ctx_mutex = _starpu_sched_ctx_get_changing_ctx_mutex(sched_ctx->id);
-			STARPU_PTHREAD_RWLOCK_RDLOCK(changing_ctx_mutex);
 			nworkers = starpu_sched_ctx_get_nworkers(sched_ctx->id);
 			if (nworkers == 0)
 				ret = -1;
 			else
 			{
+				struct _starpu_worker *worker = _starpu_get_local_worker_key();
+				if (worker)
+				{
+					STARPU_PTHREAD_MUTEX_LOCK_SCHED(&worker->sched_mutex);
+					_starpu_worker_enter_sched_op(worker);
+					STARPU_PTHREAD_MUTEX_UNLOCK_SCHED(&worker->sched_mutex);
+				}
 				_STARPU_TASK_BREAK_ON(task, push);
-				_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+				_STARPU_SCHED_BEGIN;
 				ret = sched_ctx->sched_policy->push_task(task);
-				_STARPU_TRACE_WORKER_SCHEDULING_POP;
+				_STARPU_SCHED_END;
+				if (worker)
+				{
+					STARPU_PTHREAD_MUTEX_LOCK_SCHED(&worker->sched_mutex);
+					_starpu_worker_leave_sched_op(worker);
+					STARPU_PTHREAD_MUTEX_UNLOCK_SCHED(&worker->sched_mutex);
+				}
 			}
-			STARPU_PTHREAD_RWLOCK_UNLOCK(changing_ctx_mutex);
 		}
 
 		if(ret == -1)
@@ -627,18 +669,16 @@ int _starpu_pop_task_end(struct starpu_task *task)
  * Given a handle that needs to be converted in order to be used on the given
  * node, returns a task that takes care of the conversion.
  */
-struct starpu_task *_starpu_create_conversion_task(starpu_data_handle_t handle,
-						   unsigned int node)
+struct starpu_task *_starpu_create_conversion_task(starpu_data_handle_t handle, unsigned int node)
 {
 	return _starpu_create_conversion_task_for_arch(handle, starpu_node_get_kind(node));
 }
 
-struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t handle,
-						   enum starpu_node_kind node_kind)
+struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t handle, enum starpu_node_kind node_kind)
 {
 	struct starpu_task *conversion_task;
 
-#if defined(STARPU_USE_OPENCL) || defined(STARPU_USE_CUDA) || defined(STARPU_USE_MIC) || defined(STARPU_USE_SCC) || defined(STARPU_SIMGRID)
+#if defined(STARPU_USE_OPENCL) || defined(STARPU_USE_CUDA) || defined(STARPU_USE_MIC) || defined(STARPU_SIMGRID)
 	struct starpu_multiformat_interface *format_interface;
 #endif
 
@@ -647,7 +687,7 @@ struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t
 	conversion_task->synchronous = 0;
 	STARPU_TASK_SET_HANDLE(conversion_task, handle, 0);
 
-#if defined(STARPU_USE_OPENCL) || defined(STARPU_USE_CUDA) || defined(STARPU_USE_MIC) || defined(STARPU_USE_SCC) || defined(STARPU_SIMGRID)
+#if defined(STARPU_USE_OPENCL) || defined(STARPU_USE_CUDA) || defined(STARPU_USE_MIC) || defined(STARPU_SIMGRID)
 	/* The node does not really matter here */
 	format_interface = (struct starpu_multiformat_interface *) starpu_data_get_interface_on_node(handle, STARPU_MAIN_RAM);
 #endif
@@ -660,13 +700,9 @@ struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t
 	switch(node_kind)
 	{
 	case STARPU_CPU_RAM:
-	case STARPU_SCC_RAM:
-	case STARPU_SCC_SHM:
 		switch (starpu_node_get_kind(handle->mf_node))
 		{
 		case STARPU_CPU_RAM:
-		case STARPU_SCC_RAM:
-		case STARPU_SCC_SHM:
 			STARPU_ABORT();
 #if defined(STARPU_USE_CUDA) || defined(STARPU_SIMGRID)
 		case STARPU_CUDA_RAM:
@@ -730,6 +766,7 @@ struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t
 		STARPU_ABORT();
 	}
 
+	_starpu_codelet_check_deprecated_fields(conversion_task->cl);
 	STARPU_TASK_SET_MODE(conversion_task, STARPU_RW, 0);
 	return conversion_task;
 }
@@ -737,79 +774,38 @@ struct starpu_task *_starpu_create_conversion_task_for_arch(starpu_data_handle_t
 static
 struct _starpu_sched_ctx* _get_next_sched_ctx_to_pop_into(struct _starpu_worker *worker)
 {
-	struct _starpu_sched_ctx_list *l = NULL;
-	for (l = worker->sched_ctx_list; l; l = l->next)
+	struct _starpu_sched_ctx_elt *e = NULL;
+	struct _starpu_sched_ctx_list_iterator list_it;
+	int found = 0;
+
+	_starpu_sched_ctx_list_iterator_init(worker->sched_ctx_list, &list_it);
+	while (_starpu_sched_ctx_list_iterator_has_next(&list_it))
 	{
-		if(worker->removed_from_ctx[l->sched_ctx] == 1)
-		{
-			return	_starpu_get_sched_ctx_struct(l->sched_ctx);
-		}
+		e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+		if (e->task_number > 0)
+			return _starpu_get_sched_ctx_struct(e->sched_ctx);
 	}
 
-	unsigned are_2_priorities = 0;
-	for (l = worker->sched_ctx_list; l; l = l->next)
+	_starpu_sched_ctx_list_iterator_init(worker->sched_ctx_list, &list_it);
+	while (_starpu_sched_ctx_list_iterator_has_next(&list_it))
 	{
-		if(l->priority != worker->pop_ctx_priority)
+		e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+		if (e->last_poped)
 		{
-			are_2_priorities = 1;
+			e->last_poped = 0;
+			if (_starpu_sched_ctx_list_iterator_has_next(&list_it))
+			{
+				e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+				found = 1;
+			}
 			break;
 		}
 	}
+	if (!found)
+		e = worker->sched_ctx_list->head;
+	e->last_poped = 1;
 
-	if(!worker->reverse_phase[worker->pop_ctx_priority])
-	{
-		/* find a context in which the worker hasn't poped yet */
-		for (l = worker->sched_ctx_list; l; l = l->next)
-		{
-			if(l->priority == worker->pop_ctx_priority)
-			{
-				if(!worker->poped_in_ctx[l->sched_ctx])
-				{
-					worker->poped_in_ctx[l->sched_ctx] = !worker->poped_in_ctx[l->sched_ctx];
-					return	_starpu_get_sched_ctx_struct(l->sched_ctx);
-				}
-			}
-		}
-		worker->reverse_phase[worker->pop_ctx_priority] = !worker->reverse_phase[worker->pop_ctx_priority];
-		if(are_2_priorities)
-			worker->pop_ctx_priority = !worker->pop_ctx_priority;
-	}
-	are_2_priorities = 0;
-	if(worker->reverse_phase[worker->pop_ctx_priority])
-	{
-		/* if the context has already poped in every one start from the begining */
-		for (l = worker->sched_ctx_list; l; l = l->next)
-		{
-			if(l->priority == worker->pop_ctx_priority)
-			{
-				if(worker->poped_in_ctx[l->sched_ctx])
-				{
-					worker->poped_in_ctx[l->sched_ctx] = !worker->poped_in_ctx[l->sched_ctx];
-					return	_starpu_get_sched_ctx_struct(l->sched_ctx);
-				}
-			}
-		}
-		worker->reverse_phase[worker->pop_ctx_priority] = !worker->reverse_phase[worker->pop_ctx_priority];
-		if(are_2_priorities)
-			worker->pop_ctx_priority = !worker->pop_ctx_priority;
-	}
-
-	unsigned first_sched_ctx = STARPU_NMAX_SCHED_CTXS;
-	for (l = worker->sched_ctx_list; l; l = l->next)
-	{
-		if(l->priority == worker->pop_ctx_priority)
-		{
-			first_sched_ctx = l->sched_ctx;
-			break;
-		}
-	}
-
-//	if(worker->pop_ctx_priority == 0 && first_sched_ctx == STARPU_NMAX_SCHED_CTXS)
-	if(first_sched_ctx == STARPU_NMAX_SCHED_CTXS)
-		first_sched_ctx = worker->sched_ctx_list->sched_ctx;
-
-	worker->poped_in_ctx[first_sched_ctx] = !worker->poped_in_ctx[first_sched_ctx];
-	return _starpu_get_sched_ctx_struct(first_sched_ctx);
+	return _starpu_get_sched_ctx_struct(e->sched_ctx);
 }
 
 struct starpu_task *_starpu_pop_task(struct _starpu_worker *worker)
@@ -851,6 +847,14 @@ pick:
 			{
 				while(1)
 				{
+					/** Caution
+					 * If you use multiple contexts your scheduler *needs*
+					 * to update the variable task_number of the ctx list.
+					 * In order to get the best performances.
+					 * This is done using functions :
+					 *   starpu_sched_ctx_list_task_counters_increment...(...)
+					 *   starpu_sched_ctx_list_task_counters_decrement...(...)
+					**/
 					sched_ctx = _get_next_sched_ctx_to_pop_into(worker);
 
 					if(worker->removed_from_ctx[sched_ctx->id] == 1 && worker->shares_tasks_lists[sched_ctx->id] == 1)
@@ -868,18 +872,22 @@ pick:
 			{
 				if (sched_ctx->sched_policy && sched_ctx->sched_policy->pop_task)
 				{
+					/* Note: we do not push the scheduling state here, because
+					 * otherwise when a worker is idle, we'd keep
+					 * pushing/popping a scheduling state here, while what we
+					 * want to see in the trace is a permanent idle state. */
 					task = sched_ctx->sched_policy->pop_task(sched_ctx->id);
 					if (task)
 						_STARPU_TASK_BREAK_ON(task, pop);
 					_starpu_pop_task_end(task);
 				}
 			}
-			
+
 			if(!task)
 			{
 				/* it doesn't matter if it shares tasks list or not in the scheduler,
 				   if it does not have any task to pop just get it out of here */
-				/* however if it shares a task list it will be removed as soon as he 
+				/* however if it shares a task list it will be removed as soon as he
 				  finishes this job (in handle_job_termination) */
 				if(worker->removed_from_ctx[sched_ctx->id])
 				{
@@ -898,7 +906,7 @@ pick:
 					}
 				}
 #endif //STARPU_USE_SC_HYPERVISOR
-				
+
 #ifndef STARPU_NON_BLOCKING_DRIVERS
 				if(been_here[sched_ctx->id] || worker->nsched_ctxs == 1)
 					break;
@@ -924,7 +932,7 @@ pick:
 		idle[worker->workerid] += (idle_end - idle_start[worker->workerid]);
 		idle_start[worker->workerid] = 0.0;
 	}
-	
+
 
 #ifdef STARPU_USE_SC_HYPERVISOR
 	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(task->sched_ctx);
@@ -956,7 +964,7 @@ pick:
 	 * This way, we avoid computing and executing the conversions tasks.
 	 * Here, we do not care about what implementation is used.
 	 */
-	worker_id = starpu_worker_get_id();
+	worker_id = starpu_worker_get_id_check();
 	if (!starpu_worker_can_execute_task_first_impl(worker_id, task, NULL))
 		return task;
 
@@ -966,7 +974,6 @@ pick:
 	 * We do have a task that uses multiformat handles. Let's create the
 	 * required conversion tasks.
 	 */
-	STARPU_PTHREAD_MUTEX_UNLOCK_SCHED(&worker->sched_mutex);
 	unsigned i;
 	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
 	for (i = 0; i < nbuffers; i++)
@@ -990,7 +997,6 @@ pick:
 
 	task->mf_skip = 1;
 	starpu_task_list_push_back(&worker->local_tasks, task);
-	STARPU_PTHREAD_MUTEX_LOCK_SCHED(&worker->sched_mutex);
 	goto pick;
 
 profiling:
@@ -1011,7 +1017,11 @@ profiling:
 	}
 
 	if(task->prologue_callback_pop_func)
+	{
+		_starpu_set_current_task(task);
 		task->prologue_callback_pop_func(task->prologue_callback_pop_arg);
+		_starpu_set_current_task(NULL);
+	}
 
 	return task;
 }
@@ -1022,13 +1032,13 @@ struct starpu_task *_starpu_pop_every_task(struct _starpu_sched_ctx *sched_ctx)
 	if(sched_ctx->sched_policy)
 	{
 		STARPU_ASSERT(sched_ctx->sched_policy->pop_every_task);
-		
+
 		/* TODO set profiling info */
 		if(sched_ctx->sched_policy->pop_every_task)
 		{
-			_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
+			_STARPU_SCHED_BEGIN;
 			task = sched_ctx->sched_policy->pop_every_task(sched_ctx->id);
-			_STARPU_TRACE_WORKER_SCHEDULING_POP;
+			_STARPU_SCHED_END;
 		}
 	}
 	return task;
@@ -1036,24 +1046,76 @@ struct starpu_task *_starpu_pop_every_task(struct _starpu_sched_ctx *sched_ctx)
 
 void _starpu_sched_pre_exec_hook(struct starpu_task *task)
 {
-	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(task->sched_ctx);
+	unsigned sched_ctx_id = starpu_sched_ctx_get_ctx_for_task(task);
+	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(sched_ctx_id);
 	if (sched_ctx->sched_policy && sched_ctx->sched_policy->pre_exec_hook)
 	{
-		_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
-		sched_ctx->sched_policy->pre_exec_hook(task);
-		_STARPU_TRACE_WORKER_SCHEDULING_POP;
+		_STARPU_SCHED_BEGIN;
+		sched_ctx->sched_policy->pre_exec_hook(task, sched_ctx_id);
+		_STARPU_SCHED_END;
 	}
+
+	if(!sched_ctx->sched_policy)
+	{
+		int workerid = starpu_worker_get_id();
+		struct _starpu_worker *worker =  _starpu_get_worker_struct(workerid);
+		struct _starpu_sched_ctx_list_iterator list_it;
+
+		_starpu_sched_ctx_list_iterator_init(worker->sched_ctx_list, &list_it);
+		while (_starpu_sched_ctx_list_iterator_has_next(&list_it))
+		{
+			struct _starpu_sched_ctx *other_sched_ctx;
+			struct _starpu_sched_ctx_elt *e;
+
+			e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+			other_sched_ctx = _starpu_get_sched_ctx_struct(e->sched_ctx);
+			if (other_sched_ctx != sched_ctx &&
+			    other_sched_ctx->sched_policy != NULL &&
+			    other_sched_ctx->sched_policy->pre_exec_hook)
+			{
+				_STARPU_SCHED_BEGIN;
+				other_sched_ctx->sched_policy->pre_exec_hook(task, other_sched_ctx->id);
+				_STARPU_SCHED_END;
+			}
+		}
+	}
+
 }
 
 void _starpu_sched_post_exec_hook(struct starpu_task *task)
 {
-	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(task->sched_ctx);
-
+	STARPU_ASSERT(task->cl != NULL && task->cl->where != STARPU_NOWHERE);
+	unsigned sched_ctx_id = starpu_sched_ctx_get_ctx_for_task(task);
+	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(sched_ctx_id);
 	if (sched_ctx->sched_policy && sched_ctx->sched_policy->post_exec_hook)
 	{
-		_STARPU_TRACE_WORKER_SCHEDULING_PUSH;
-		sched_ctx->sched_policy->post_exec_hook(task);
-		_STARPU_TRACE_WORKER_SCHEDULING_POP;
+		_STARPU_SCHED_BEGIN;
+		sched_ctx->sched_policy->post_exec_hook(task, sched_ctx_id);
+		_STARPU_SCHED_END;
+	}
+	if(!sched_ctx->sched_policy)
+	{
+		int workerid = starpu_worker_get_id();
+		struct _starpu_worker *worker =  _starpu_get_worker_struct(workerid);
+		struct _starpu_sched_ctx_list_iterator list_it;
+
+		_starpu_sched_ctx_list_iterator_init(worker->sched_ctx_list, &list_it);
+		while (_starpu_sched_ctx_list_iterator_has_next(&list_it))
+		{
+			struct _starpu_sched_ctx *other_sched_ctx;
+			struct _starpu_sched_ctx_elt *e;
+
+			e = _starpu_sched_ctx_list_iterator_get_next(&list_it);
+			other_sched_ctx = _starpu_get_sched_ctx_struct(e->sched_ctx);
+			if (other_sched_ctx != sched_ctx &&
+			    other_sched_ctx->sched_policy != NULL &&
+			    other_sched_ctx->sched_policy->post_exec_hook)
+			{
+				_STARPU_SCHED_BEGIN;
+				other_sched_ctx->sched_policy->post_exec_hook(task, other_sched_ctx->id);
+				_STARPU_SCHED_END;
+			}
+		}
 	}
 }
 
