@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2011-2012                                Inria
- * Copyright (C) 2008-2016                                Université de Bordeaux
- * Copyright (C) 2010-2017                                CNRS
+ * Copyright (C) 2008-2019                                Université de Bordeaux
+ * Copyright (C) 2011,2012,2017                           Inria
+ * Copyright (C) 2010-2017,2019                           CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -17,17 +17,6 @@
  */
 
 #include <starpu.h>
-#include <common/config.h>
-#include <datawizard/coherency.h>
-#include <datawizard/copy_driver.h>
-#include <datawizard/filters.h>
-#include <datawizard/memory_nodes.h>
-#include <starpu_hash.h>
-#include <starpu_cuda.h>
-#include <starpu_opencl.h>
-#include <drivers/opencl/driver_opencl.h>
-#include <drivers/scc/driver_scc_source.h>
-#include <drivers/mic/driver_mic_source.h>
 
 #ifdef STARPU_USE_CUDA
 /* At least CUDA 4.2 still didn't have working memcpy3D */
@@ -84,13 +73,18 @@ static const struct starpu_data_copy_methods matrix_copy_data_methods_s =
 	.any_to_any = copy_any_to_any,
 };
 
+static void matrix_init(void *data_interface);
 static void register_matrix_handle(starpu_data_handle_t handle, unsigned home_node, void *data_interface);
-static void *matrix_handle_to_pointer(starpu_data_handle_t data_handle, unsigned node);
+static void *matrix_to_pointer(void *data_interface, unsigned node);
+static int matrix_pointer_is_inside(void *data_interface, unsigned node, void *ptr);
 static starpu_ssize_t allocate_matrix_buffer_on_node(void *data_interface_, unsigned dst_node);
 static void free_matrix_buffer_on_node(void *data_interface, unsigned node);
 static size_t matrix_interface_get_size(starpu_data_handle_t handle);
+static size_t matrix_interface_get_alloc_size(starpu_data_handle_t handle);
 static uint32_t footprint_matrix_interface_crc32(starpu_data_handle_t handle);
+static uint32_t alloc_footprint_matrix_interface_crc32(starpu_data_handle_t handle);
 static int matrix_compare(void *data_interface_a, void *data_interface_b);
+static int matrix_alloc_compare(void *data_interface_a, void *data_interface_b);
 static void display_matrix_interface(starpu_data_handle_t handle, FILE *f);
 static int pack_matrix_handle(starpu_data_handle_t handle, unsigned node, void **ptr, starpu_ssize_t *count);
 static int unpack_matrix_handle(starpu_data_handle_t handle, unsigned node, void *ptr, size_t count);
@@ -98,14 +92,19 @@ static starpu_ssize_t describe(void *data_interface, char *buf, size_t size);
 
 struct starpu_data_interface_ops starpu_interface_matrix_ops =
 {
+	.init = matrix_init,
 	.register_data_handle = register_matrix_handle,
 	.allocate_data_on_node = allocate_matrix_buffer_on_node,
-	.handle_to_pointer = matrix_handle_to_pointer,
+	.to_pointer = matrix_to_pointer,
+	.pointer_is_inside = matrix_pointer_is_inside,
 	.free_data_on_node = free_matrix_buffer_on_node,
 	.copy_methods = &matrix_copy_data_methods_s,
 	.get_size = matrix_interface_get_size,
+	.get_alloc_size = matrix_interface_get_alloc_size,
 	.footprint = footprint_matrix_interface_crc32,
+	.alloc_footprint = alloc_footprint_matrix_interface_crc32,
 	.compare = matrix_compare,
+	.alloc_compare = matrix_alloc_compare,
 	.interfaceid = STARPU_MATRIX_INTERFACE_ID,
 	.interface_size = sizeof(struct starpu_matrix_interface),
 	.display = display_matrix_interface,
@@ -115,6 +114,12 @@ struct starpu_data_interface_ops starpu_interface_matrix_ops =
 	.name = "STARPU_MATRIX_INTERFACE"
 };
 
+static void matrix_init(void *data_interface)
+{
+	struct starpu_matrix_interface *matrix_interface = data_interface;
+	matrix_interface->allocsize = -1;
+}
+
 static void register_matrix_handle(starpu_data_handle_t handle, unsigned home_node, void *data_interface)
 {
 	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *) data_interface;
@@ -144,24 +149,32 @@ static void register_matrix_handle(starpu_data_handle_t handle, unsigned home_no
 		local_interface->nx = matrix_interface->nx;
 		local_interface->ny = matrix_interface->ny;
 		local_interface->elemsize = matrix_interface->elemsize;
+		local_interface->allocsize  = matrix_interface->allocsize;
 	}
 }
 
-static void *matrix_handle_to_pointer(starpu_data_handle_t handle, unsigned node)
+static void *matrix_to_pointer(void *data_interface, unsigned node)
 {
-	STARPU_ASSERT(starpu_data_test_if_allocated_on_node(handle, node));
-
-	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *)
-		starpu_data_get_interface_on_node(handle, node);
+	(void) node;
+	struct starpu_matrix_interface *matrix_interface = data_interface;
 
 	return (void*) matrix_interface->ptr;
 }
 
+static int matrix_pointer_is_inside(void *data_interface, unsigned node, void *ptr)
+{
+	(void) node;
+	struct starpu_matrix_interface *matrix_interface = data_interface;
+
+	return (char*) ptr >= (char*) matrix_interface->ptr &&
+		(char*) ptr < (char*) matrix_interface->ptr + matrix_interface->nx*matrix_interface->ny*matrix_interface->elemsize;
+}
+
 
 /* declare a new data with the matrix interface */
-void starpu_matrix_data_register(starpu_data_handle_t *handleptr, int home_node,
-			uintptr_t ptr, uint32_t ld, uint32_t nx,
-			uint32_t ny, size_t elemsize)
+void starpu_matrix_data_register_allocsize(starpu_data_handle_t *handleptr, int home_node,
+					   uintptr_t ptr, uint32_t ld, uint32_t nx,
+					   uint32_t ny, size_t elemsize, size_t allocsize)
 {
 	struct starpu_matrix_interface matrix_interface =
 	{
@@ -172,26 +185,29 @@ void starpu_matrix_data_register(starpu_data_handle_t *handleptr, int home_node,
 		.ny = ny,
 		.elemsize = elemsize,
                 .dev_handle = ptr,
-                .offset = 0
+                .offset = 0,
+		.allocsize = allocsize,
 	};
 #ifndef STARPU_SIMGRID
-	if (home_node == STARPU_MAIN_RAM)
+	if (home_node >= 0 && starpu_node_get_kind(home_node) == STARPU_CPU_RAM)
 	{
 		STARPU_ASSERT_ACCESSIBLE(ptr);
 		STARPU_ASSERT_ACCESSIBLE(ptr + (ny-1)*ld*elemsize + nx*elemsize - 1);
 	}
 #endif
 
-#ifdef STARPU_USE_SCC
-	_starpu_scc_set_offset_in_shared_memory((void*)matrix_interface.ptr,
-			(void**)&(matrix_interface.dev_handle), &(matrix_interface.offset));
-#endif
-
 	starpu_data_register(handleptr, home_node, &matrix_interface, &starpu_interface_matrix_ops);
 }
 
+void starpu_matrix_data_register(starpu_data_handle_t *handleptr, int home_node,
+				 uintptr_t ptr, uint32_t ld, uint32_t nx,
+				 uint32_t ny, size_t elemsize)
+{
+	starpu_matrix_data_register_allocsize(handleptr, home_node, ptr, ld, nx, ny, elemsize, nx * ny * elemsize);
+}
+
 void starpu_matrix_ptr_register(starpu_data_handle_t handle, unsigned node,
-			uintptr_t ptr, uintptr_t dev_handle, size_t offset, uint32_t ld)
+				uintptr_t ptr, uintptr_t dev_handle, size_t offset, uint32_t ld)
 {
 	struct starpu_matrix_interface *matrix_interface = starpu_data_get_interface_on_node(handle, node);
 	starpu_data_ptr_register(handle, node);
@@ -206,6 +222,11 @@ static uint32_t footprint_matrix_interface_crc32(starpu_data_handle_t handle)
 	return starpu_hash_crc32c_be(starpu_matrix_get_nx(handle), starpu_matrix_get_ny(handle));
 }
 
+static uint32_t alloc_footprint_matrix_interface_crc32(starpu_data_handle_t handle)
+{
+	return starpu_hash_crc32c_be(starpu_matrix_get_allocsize(handle), 0);
+}
+
 static int matrix_compare(void *data_interface_a, void *data_interface_b)
 {
 	struct starpu_matrix_interface *matrix_a = (struct starpu_matrix_interface *) data_interface_a;
@@ -217,6 +238,15 @@ static int matrix_compare(void *data_interface_a, void *data_interface_b)
 		&& (matrix_a->elemsize == matrix_b->elemsize);
 }
 
+static int matrix_alloc_compare(void *data_interface_a, void *data_interface_b)
+{
+	struct starpu_matrix_interface *matrix_a = (struct starpu_matrix_interface *) data_interface_a;
+	struct starpu_matrix_interface *matrix_b = (struct starpu_matrix_interface *) data_interface_b;
+
+	/* Two matricess are considered compatible if they have the same size */
+	return (matrix_a->allocsize == matrix_b->allocsize);
+}
+
 static void display_matrix_interface(starpu_data_handle_t handle, FILE *f)
 {
 	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *)
@@ -239,7 +269,7 @@ static int pack_matrix_handle(starpu_data_handle_t handle, unsigned node, void *
 		uint32_t y;
 		char *matrix = (void *)matrix_interface->ptr;
 
-		starpu_malloc_flags(ptr, *count, 0);
+		*ptr = (void *)starpu_malloc_on_node_flags(node, *count, 0);
 
 		char *cur = *ptr;
 		for(y=0 ; y<matrix_interface->ny ; y++)
@@ -272,6 +302,8 @@ static int unpack_matrix_handle(starpu_data_handle_t handle, unsigned node, void
 		matrix += matrix_interface->ld * matrix_interface->elemsize;
 	}
 
+	starpu_free_on_node_flags(node, (uintptr_t)ptr, count, 0);
+
 	return 0;
 }
 
@@ -284,10 +316,21 @@ static size_t matrix_interface_get_size(starpu_data_handle_t handle)
 	STARPU_ASSERT_MSG(matrix_interface->id == STARPU_MATRIX_INTERFACE_ID, "Error. The given data is not a matrix.");
 #endif
 
-	size_t size;
-	size = (size_t)matrix_interface->nx*matrix_interface->ny*matrix_interface->elemsize;
+	return matrix_interface->nx * matrix_interface->ny * matrix_interface->elemsize;
+}
 
-	return size;
+static size_t matrix_interface_get_alloc_size(starpu_data_handle_t handle)
+{
+	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *)
+		starpu_data_get_interface_on_node(handle, STARPU_MAIN_RAM);
+
+#ifdef STARPU_DEBUG
+	STARPU_ASSERT_MSG(matrix_interface->id == STARPU_MATRIX_INTERFACE_ID, "Error. The given data is not a matrix.");
+#endif
+
+	STARPU_ASSERT_MSG(matrix_interface->allocsize != (size_t)-1, "The matrix allocation size needs to be defined");
+
+	return matrix_interface->allocsize;
 }
 
 /* offer an access to the data parameters */
@@ -318,7 +361,7 @@ uint32_t starpu_matrix_get_ny(starpu_data_handle_t handle)
 uint32_t starpu_matrix_get_local_ld(starpu_data_handle_t handle)
 {
 	unsigned node;
-	node = _starpu_memory_node_get_local_key();
+	node = starpu_worker_get_local_memory_node();
 
 	STARPU_ASSERT(starpu_data_test_if_allocated_on_node(handle, node));
 
@@ -335,7 +378,7 @@ uint32_t starpu_matrix_get_local_ld(starpu_data_handle_t handle)
 uintptr_t starpu_matrix_get_local_ptr(starpu_data_handle_t handle)
 {
 	unsigned node;
-	node = _starpu_memory_node_get_local_key();
+	node = starpu_worker_get_local_memory_node();
 
 	STARPU_ASSERT(starpu_data_test_if_allocated_on_node(handle, node));
 
@@ -361,6 +404,18 @@ size_t starpu_matrix_get_elemsize(starpu_data_handle_t handle)
 	return matrix_interface->elemsize;
 }
 
+size_t starpu_matrix_get_allocsize(starpu_data_handle_t handle)
+{
+	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *)
+		starpu_data_get_interface_on_node(handle, STARPU_MAIN_RAM);
+
+#ifdef STARPU_DEBUG
+	STARPU_ASSERT_MSG(matrix_interface->id == STARPU_MATRIX_INTERFACE_ID, "Error. The given data is not a matrix.");
+#endif
+
+	return matrix_interface->allocsize;
+}
+
 /* memory allocation/deallocation primitives for the matrix interface */
 
 /* returns the size of the allocated area */
@@ -370,14 +425,10 @@ static starpu_ssize_t allocate_matrix_buffer_on_node(void *data_interface_, unsi
 
 	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *) data_interface_;
 
-	uint32_t nx = matrix_interface->nx;
-	uint32_t ny = matrix_interface->ny;
-	uint32_t ld = nx; // by default
-	size_t elemsize = matrix_interface->elemsize;
+	uint32_t ld = matrix_interface->nx; // by default
 
-	starpu_ssize_t allocated_memory;
-
-	handle = starpu_malloc_on_node(dst_node, nx*ny*elemsize);
+	starpu_ssize_t allocated_memory = matrix_interface->allocsize;
+	handle = starpu_malloc_on_node(dst_node, allocated_memory);
 
 	if (!handle)
 		return -ENOMEM;
@@ -385,8 +436,6 @@ static starpu_ssize_t allocate_matrix_buffer_on_node(void *data_interface_, unsi
 	if (starpu_node_get_kind(dst_node) != STARPU_OPENCL_RAM)
 		addr = handle;
 
-	allocated_memory = (size_t)nx*ny*elemsize;
-
 	/* update the data properly in consequence */
 	matrix_interface->ptr = addr;
 	matrix_interface->dev_handle = handle;
@@ -399,11 +448,8 @@ static starpu_ssize_t allocate_matrix_buffer_on_node(void *data_interface_, unsi
 static void free_matrix_buffer_on_node(void *data_interface, unsigned node)
 {
 	struct starpu_matrix_interface *matrix_interface = (struct starpu_matrix_interface *) data_interface;
-	uint32_t nx = matrix_interface->nx;
-	uint32_t ny = matrix_interface->ny;
-	size_t elemsize = matrix_interface->elemsize;
 
-	starpu_free_on_node(node, matrix_interface->dev_handle, nx*ny*elemsize);
+	starpu_free_on_node(node, matrix_interface->dev_handle, matrix_interface->allocsize);
 }
 
 #ifdef STARPU_USE_CUDA
@@ -417,11 +463,12 @@ static int copy_cuda_common(void *src_interface, unsigned src_node STARPU_ATTRIB
 
 	if (is_async)
 	{
-		_STARPU_TRACE_START_DRIVER_COPY_ASYNC(src_node, dst_node);
+		double start;
+		starpu_interface_start_driver_copy_async(src_node, dst_node, &start);
 		cures = cudaMemcpy2DAsync((char *)dst_matrix->ptr, dst_matrix->ld*elemsize,
 			(char *)src_matrix->ptr, src_matrix->ld*elemsize,
 			src_matrix->nx*elemsize, src_matrix->ny, kind, stream);
-		_STARPU_TRACE_END_DRIVER_COPY_ASYNC(src_node, dst_node);
+		starpu_interface_end_driver_copy_async(src_node, dst_node, start);
 		if (!cures)
 			return -EAGAIN;
 	}
@@ -429,6 +476,8 @@ static int copy_cuda_common(void *src_interface, unsigned src_node STARPU_ATTRIB
 	cures = cudaMemcpy2D((char *)dst_matrix->ptr, dst_matrix->ld*elemsize,
 		(char *)src_matrix->ptr, src_matrix->ld*elemsize,
 		src_matrix->nx*elemsize, src_matrix->ny, kind);
+	if (!cures)
+		cures = cudaThreadSynchronize();
 	if (STARPU_UNLIKELY(cures))
 	{
 		int ret = copy_any_to_any(src_interface, src_node, dst_interface, dst_node, (void*)(uintptr_t)is_async);
@@ -436,7 +485,7 @@ static int copy_cuda_common(void *src_interface, unsigned src_node STARPU_ATTRIB
 		if (ret) STARPU_CUDA_REPORT_ERROR(cures);
 	}
 
-	_STARPU_TRACE_DATA_COPY(src_node, dst_node, (size_t)src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
+	starpu_interface_data_copy(src_node, dst_node, (size_t)src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
 
 	return 0;
 }
@@ -451,8 +500,8 @@ static int copy_cuda_peer(void *src_interface, unsigned src_node STARPU_ATTRIBUT
 	size_t elemsize = src_matrix->elemsize;
 	cudaError_t cures;
 
-	int src_dev = _starpu_memory_node_get_devid(src_node);
-	int dst_dev = _starpu_memory_node_get_devid(dst_node);
+	int src_dev = starpu_memory_node_get_devid(src_node);
+	int dst_dev = starpu_memory_node_get_devid(dst_node);
 
 	struct cudaMemcpy3DPeerParms p;
 	memset(&p, 0, sizeof(p));
@@ -465,18 +514,21 @@ static int copy_cuda_peer(void *src_interface, unsigned src_node STARPU_ATTRIBUT
 
 	if (is_async)
 	{
-		_STARPU_TRACE_START_DRIVER_COPY_ASYNC(src_node, dst_node);
+		double start;
+		starpu_interface_start_driver_copy_async(src_node, dst_node, &start);
 		cures = cudaMemcpy3DPeerAsync(&p, stream);
-		_STARPU_TRACE_END_DRIVER_COPY_ASYNC(src_node, dst_node);
+		starpu_interface_end_driver_copy_async(src_node, dst_node, start);
 		if (!cures)
 			return -EAGAIN;
 	}
 
 	cures = cudaMemcpy3DPeer(&p);
+	if (!cures)
+		cures = cudaThreadSynchronize();
 	if (STARPU_UNLIKELY(cures))
 		STARPU_CUDA_REPORT_ERROR(cures);
 
-	_STARPU_TRACE_DATA_COPY(src_node, dst_node, (size_t)src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
+	starpu_interface_data_copy(src_node, dst_node, (size_t)src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
 
 	return 0;
 #else
@@ -542,7 +594,7 @@ static int copy_opencl_common(void *src_interface, unsigned src_node, void *dst_
 					    src_matrix->nx*src_matrix->ny*src_matrix->elemsize,
 					    event);
 
-	_STARPU_TRACE_DATA_COPY(src_node, dst_node, src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
+	starpu_interface_data_copy(src_node, dst_node, src_matrix->nx*src_matrix->ny*src_matrix->elemsize);
 
 	return ret;
 }
@@ -615,7 +667,7 @@ static int copy_any_to_any(void *src_interface, unsigned src_node, void *dst_int
 		}
 	}
 
-	_STARPU_TRACE_DATA_COPY(src_node, dst_node, (size_t)nx*ny*elemsize);
+	starpu_interface_data_copy(src_node, dst_node, (size_t)nx*ny*elemsize);
 
 	return ret;
 }
