@@ -1,7 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009-2012, 2014-2016  Université de Bordeaux
- * Copyright (C) 2010, 2011, 2013  CNRS
+ * Copyright (C) 2009-2012,2014-2017,2019                 Université de Bordeaux
+ * Copyright (C) 2012,2016,2017                           Inria
+ * Copyright (C) 2010,2011,2013,2015,2017,2019            CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -22,6 +23,7 @@
 #include <common/config.h>
 #include <datawizard/coherency.h>
 #include <datawizard/memalloc.h>
+#include <datawizard/node_ops.h>
 #include <common/utils.h>
 #include <core/workers.h>
 
@@ -29,20 +31,19 @@
 #include <core/simgrid.h>
 #endif
 
-#define _STARPU_MEMORY_NODE_TUPLE(node1,node2) (node1 | (node2 << 4))
-#define _STARPU_MEMORY_NODE_TUPLE_FIRST(tuple) (tuple & 0x0F)
-#define _STARPU_MEMORY_NODE_TUPLE_SECOND(tuple) (tuple & 0xF0)
+extern char _starpu_worker_drives_memory[STARPU_NMAXWORKERS][STARPU_MAXNODES];
 
-struct _starpu_cond_and_mutex
+struct _starpu_cond_and_worker
 {
-        starpu_pthread_cond_t *cond;
-        starpu_pthread_mutex_t *mutex;
+	starpu_pthread_cond_t *cond;
+	struct _starpu_worker *worker;
 };
 
 struct _starpu_memory_node_descr
 {
 	unsigned nnodes;
 	enum starpu_node_kind nodes[STARPU_MAXNODES];
+	struct _starpu_node_ops *node_ops[STARPU_MAXNODES];
 
 	/* Get the device id associated to this node, or -1 if not applicable */
 	int devid[STARPU_MAXNODES];
@@ -60,42 +61,31 @@ struct _starpu_memory_node_descr
 	 * list of all these condition variables so that we can wake up all
 	 * worker attached to a memory node that are waiting on a task. */
 	starpu_pthread_rwlock_t conditions_rwlock;
-	struct _starpu_cond_and_mutex conditions_attached_to_node[STARPU_MAXNODES][STARPU_NMAXWORKERS];
-	struct _starpu_cond_and_mutex conditions_all[STARPU_MAXNODES*STARPU_NMAXWORKERS];
+	struct _starpu_cond_and_worker conditions_attached_to_node[STARPU_MAXNODES][STARPU_NMAXWORKERS];
+	struct _starpu_cond_and_worker conditions_all[STARPU_MAXNODES*STARPU_NMAXWORKERS];
 	/* the number of queues attached to each node */
 	unsigned total_condition_count;
 	unsigned condition_count[STARPU_MAXNODES];
-
 };
 
 extern struct _starpu_memory_node_descr _starpu_descr;
 
 void _starpu_memory_nodes_init(void);
 void _starpu_memory_nodes_deinit(void);
-extern starpu_pthread_key_t _starpu_memory_node_key STARPU_ATTRIBUTE_INTERNAL;
-static inline void _starpu_memory_node_set_local_key(unsigned *node)
-{
-	STARPU_PTHREAD_SETSPECIFIC(_starpu_memory_node_key, node);
-}
-
-static inline unsigned _starpu_memory_node_get_local_key(void)
-{
-	unsigned *memory_node;
-	memory_node = (unsigned *) STARPU_PTHREAD_GETSPECIFIC(_starpu_memory_node_key);
-
-	/* in case this is called by the programmer, we assume the RAM node
-	   is the appropriate memory node ... XXX */
-	if (STARPU_UNLIKELY(!memory_node))
-		return STARPU_MAIN_RAM;
-
-	return *memory_node;
-}
 
 static inline void _starpu_memory_node_add_nworkers(unsigned node)
 {
 	_starpu_descr.nworkers[node]++;
 }
 
+/* same utility as _starpu_memory_node_add_nworkers */
+void _starpu_worker_drives_memory_node(struct _starpu_worker *worker, unsigned memnode);
+
+static inline struct _starpu_node_ops *_starpu_memory_node_get_node_ops(unsigned node)
+{
+	return _starpu_descr.node_ops[node];
+}
+
 static inline unsigned _starpu_memory_node_get_nworkers(unsigned node)
 {
 	return _starpu_descr.nworkers[node];
@@ -112,16 +102,9 @@ static inline msg_host_t _starpu_simgrid_memory_node_get_host(unsigned node)
 	return _starpu_descr.host[node];
 }
 #endif
-unsigned _starpu_memory_node_register(enum starpu_node_kind kind, int devid);
+unsigned _starpu_memory_node_register(enum starpu_node_kind kind, int devid, struct _starpu_node_ops *node_ops);
 //void _starpu_memory_node_attach_queue(struct starpu_jobq_s *q, unsigned nodeid);
-void _starpu_memory_node_register_condition(starpu_pthread_cond_t *cond, starpu_pthread_mutex_t *mutex, unsigned memory_node);
-
-static inline int _starpu_memory_node_get_devid(unsigned node)
-{
-	return _starpu_descr.devid[node];
-}
-
-void _starpu_memory_node_get_name(unsigned node, char *name, int size);
+void _starpu_memory_node_register_condition(struct _starpu_worker *worker, starpu_pthread_cond_t *cond, unsigned nodeid);
 
 static inline struct _starpu_memory_node_descr *_starpu_memory_node_get_description(void)
 {
