@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2013                                     Inria
- * Copyright (C) 2014-2015,2017                           CNRS
+ * Copyright (C) 2013,2017                                Inria
+ * Copyright (C) 2014,2015,2017                           CNRS
  * Copyright (C) 2013-2018                                Université de Bordeaux
  * Copyright (C) 2013                                     Simon Archipoff
  *
@@ -22,11 +22,12 @@
 #include "helper_mct.h"
 #include <float.h>
 #include <core/sched_policy.h>
+#include <core/task.h>
 
 static int mct_push_task(struct starpu_sched_component * component, struct starpu_task * task)
 {
 	STARPU_ASSERT(component && task && starpu_sched_component_is_mct(component));
-	struct _starpu_mct_data * d = component->data;	
+	struct _starpu_mct_data * d = component->data;
 	struct starpu_sched_component * best_component = NULL;
 
 	/* Estimated task duration for each child */
@@ -36,7 +37,7 @@ static int mct_push_task(struct starpu_sched_component * component, struct starp
 	/* Estimated transfer+task termination for each child */
 	double estimated_ends_with_task[component->nchildren];
 
-	int i;
+	unsigned i;
 	for(i=0; i < component->nchildren; i++)
 	{
 		estimated_lengths[i] = 0.0;
@@ -49,12 +50,11 @@ static int mct_push_task(struct starpu_sched_component * component, struct starp
 	/* Maximum transfer+task termination on all children */
 	double max_exp_end_with_task = 0.0;
 
-	int suitable_components[component->nchildren];
-	int nsuitable_components = 0;
+	unsigned suitable_components[component->nchildren];
+	unsigned nsuitable_components = 0;
 
-	nsuitable_components = starpu_mct_compute_expected_times(component, task,
-			estimated_lengths, estimated_transfer_length, estimated_ends_with_task,
-			&min_exp_end_with_task, &max_exp_end_with_task, suitable_components);
+	nsuitable_components = starpu_mct_compute_execution_times(component, task,
+								  estimated_lengths, estimated_transfer_length, suitable_components);
 
 	/* If no suitable components were found, it means that the perfmodel of
 	 * the task had been purged since it has been pushed on the mct component.
@@ -63,6 +63,15 @@ static int mct_push_task(struct starpu_sched_component * component, struct starp
 	if(nsuitable_components == 0)
 		return 1;
 
+
+	/* Entering critical section to make sure no two workers
+	   make scheduling decisions at the same time */
+	STARPU_COMPONENT_MUTEX_LOCK(&d->scheduling_mutex);
+
+
+	starpu_mct_compute_expected_times(component, task, estimated_lengths, estimated_transfer_length,
+					  estimated_ends_with_task, &min_exp_end_with_task, &max_exp_end_with_task, suitable_components, nsuitable_components);
+
 	double best_fitness = DBL_MAX;
 	int best_icomponent = -1;
 	for(i = 0; i < nsuitable_components; i++)
@@ -72,7 +81,7 @@ static int mct_push_task(struct starpu_sched_component * component, struct starp
 #warning FIXME: take energy consumption into account
 #endif
 		double tmp = starpu_mct_compute_fitness(d,
-					     estimated_ends_with_task[icomponent] - estimated_transfer_length[icomponent],
+					     estimated_ends_with_task[icomponent],
 					     min_exp_end_with_task,
 					     max_exp_end_with_task,
 					     estimated_transfer_length[icomponent],
@@ -90,18 +99,31 @@ static int mct_push_task(struct starpu_sched_component * component, struct starp
 	 * We should send a push_fail message to its parent so that it will
 	 * be able to reschedule the task properly. */
 	if(best_icomponent == -1)
+	{
+		STARPU_COMPONENT_MUTEX_UNLOCK(&d->scheduling_mutex);
 		return 1;
+	}
 
 	best_component = component->children[best_icomponent];
 
+	task->predicted = estimated_lengths[best_icomponent];
+	task->predicted_transfer = estimated_transfer_length[best_icomponent];
+
 	if(starpu_sched_component_is_worker(best_component))
 	{
 		best_component->can_pull(best_component);
+		STARPU_COMPONENT_MUTEX_UNLOCK(&d->scheduling_mutex);
+
 		return 1;
 	}
 
 	starpu_sched_task_break(task);
 	int ret = starpu_sched_component_push_task(component, best_component, task);
+
+	/* I can now exit the critical section: Pushing the task below ensures that its execution
+	   time will be taken into account for subsequent scheduling decisions */
+	STARPU_COMPONENT_MUTEX_UNLOCK(&d->scheduling_mutex);
+
 	return ret;
 }
 
@@ -109,6 +131,7 @@ static void mct_component_deinit_data(struct starpu_sched_component * component)
 {
 	STARPU_ASSERT(starpu_sched_component_is_mct(component));
 	struct _starpu_mct_data * d = component->data;
+	STARPU_PTHREAD_MUTEX_DESTROY(&d->scheduling_mutex);
 	free(d);
 }
 
@@ -123,6 +146,7 @@ struct starpu_sched_component * starpu_sched_component_mct_create(struct starpu_
 	struct _starpu_mct_data *data = starpu_mct_init_parameters(params);
 
 	component->data = data;
+	STARPU_PTHREAD_MUTEX_INIT(&data->scheduling_mutex, NULL);
 
 	component->push_task = mct_push_task;
 	component->deinit_data = mct_component_deinit_data;
