@@ -1,8 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2013-2014                                Inria
- * Copyright (C) 2014-2015,2017                           CNRS
- * Copyright (C) 2014,2017                                Université de Bordeaux
+ * Copyright (C) 2013-2015,2017                           Inria
+ * Copyright (C) 2014,2017                                CNRS
+ * Copyright (C) 2014,2017,2018                           Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -20,16 +20,14 @@
 #include <starpu_scheduler.h>
 #include <limits.h>
 
-#define _STARPU_SCHED_NTASKS_THRESHOLD_DEFAULT 4
-#define _STARPU_SCHED_EXP_LEN_THRESHOLD_DEFAULT 1000000000.0
-
+/* Just as documentation example, here is the detailed equivalent of the
+ * starpu_sched_component_initialize_simple_scheduler call below */
+#if 0
 static void initialize_prio_prefetching_center_policy(unsigned sched_ctx_id)
 {
 	struct starpu_sched_tree *t;
 	struct starpu_sched_component * eager_component;
 
-	starpu_sched_ctx_create_worker_collection(sched_ctx_id, STARPU_WORKER_LIST);
-
 	t = starpu_sched_tree_create(sched_ctx_id);
  	t->root = starpu_sched_component_prio_create(t, NULL);
 	eager_component = starpu_sched_component_eager_create(t, NULL);
@@ -45,7 +43,7 @@ static void initialize_prio_prefetching_center_policy(unsigned sched_ctx_id)
 	unsigned i;
 	for(i = 0; i < starpu_worker_get_count() + starpu_combined_worker_get_count(); i++)
 	{
-		struct starpu_sched_component * worker_component = starpu_sched_component_worker_get(sched_ctx_id, i);
+		struct starpu_sched_component * worker_component = starpu_sched_component_worker_new(sched_ctx_id, i);
 		struct starpu_sched_component * prio_component = starpu_sched_component_prio_create(t, &prio_data);
 
 		starpu_sched_component_connect(prio_component, worker_component);
@@ -60,12 +58,22 @@ static void initialize_prio_prefetching_center_policy(unsigned sched_ctx_id)
 	if (starpu_sched_ctx_max_priority_is_set(sched_ctx_id) == 0)
 		starpu_sched_ctx_set_max_priority(sched_ctx_id, INT_MAX);
 }
+#endif
+
+static void initialize_prio_prefetching_center_policy(unsigned sched_ctx_id)
+{
+	starpu_sched_component_initialize_simple_scheduler((starpu_sched_component_create_t) starpu_sched_component_eager_create, NULL,
+			STARPU_SCHED_SIMPLE_DECIDE_WORKERS |
+			STARPU_SCHED_SIMPLE_FIFO_ABOVE |
+			STARPU_SCHED_SIMPLE_FIFO_ABOVE_PRIO |
+			STARPU_SCHED_SIMPLE_FIFOS_BELOW |
+			STARPU_SCHED_SIMPLE_IMPL, sched_ctx_id);
+}
 
 static void deinitialize_prio_prefetching_center_policy(unsigned sched_ctx_id)
 {
 	struct starpu_sched_tree *tree = (struct starpu_sched_tree*)starpu_sched_ctx_get_policy_data(sched_ctx_id);
 	starpu_sched_tree_destroy(tree);
-	starpu_sched_ctx_delete_worker_collection(sched_ctx_id);
 }
 
 struct starpu_sched_policy _starpu_sched_modular_prio_prefetching_policy =
@@ -80,5 +88,6 @@ struct starpu_sched_policy _starpu_sched_modular_prio_prefetching_policy =
 	.post_exec_hook = starpu_sched_component_worker_post_exec_hook,
 	.pop_every_task = NULL,
 	.policy_name = "modular-prio-prefetching",
-	.policy_description = "prio prefetching modular policy"
+	.policy_description = "prio prefetching modular policy",
+	.worker_type = STARPU_WORKER_LIST,
 };
