@@ -1,7 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2010  Université de Bordeaux 1
- * Copyright (C) 2010, 2011, 2012  Centre National de la Recherche Scientifique
+ * Copyright (C) 2010,2011,2013,2014,2016                 Université de Bordeaux
+ * Copyright (C) 2012,2013                                Inria
+ * Copyright (C) 2010-2013,2015,2017                      CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -15,23 +16,22 @@
  * See the GNU Lesser General Public License in COPYING.LGPL for more details.
  */
 
-#include <config.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <errno.h>
 #include <starpu.h>
 
-#ifdef STARPU_USE_GORDON
-#include <gordon.h>
-#endif
-
 #include "../helper.h"
 
+/*
+ * Mix tasks with implicit dependencies and data acquisitions
+ */
+
 #define N_DEF	100
 #define K_DEF	256
 
-static int n=N_DEF;
-static int k=K_DEF;
+static unsigned n=N_DEF;
+static unsigned k=K_DEF;
 
 /*
  * In this test, we maintain a vector v = (a,b,c).
@@ -49,34 +49,31 @@ static int k=K_DEF;
  */
 
 #ifdef STARPU_USE_CUDA
-void cuda_codelet_incA(void *descr[], __attribute__ ((unused)) void *_args);
-void cuda_codelet_incC(void *descr[], __attribute__ ((unused)) void *_args);
+void cuda_codelet_incA(void *descr[], void *_args);
+void cuda_codelet_incC(void *descr[], void *_args);
 #endif
 
 #ifdef STARPU_USE_OPENCL
-#include <starpu_opencl.h>
-void opencl_codelet_incA(void *descr[], __attribute__ ((unused)) void *_args);
-void opencl_codelet_incC(void *descr[], __attribute__ ((unused)) void *_args);
+void opencl_codelet_incA(void *descr[], void *_args);
+void opencl_codelet_incC(void *descr[], void *_args);
 struct starpu_opencl_program opencl_code;
 #endif
 
 #define VECTORSIZE	16
 
 starpu_data_handle_t v_handle;
-static unsigned v[VECTORSIZE] __attribute__((aligned(128))) = {0, 0, 0, 0};
+static unsigned v[VECTORSIZE] STARPU_ATTRIBUTE_ALIGNED(128) = {0, 0, 0, 0};
 
-void cpu_codelet_incA(void *descr[], __attribute__ ((unused)) void *_args)
+void cpu_codelet_incA(void *descr[], void *arg)
 {
-	STARPU_SKIP_IF_VALGRIND;
-
+	(void)arg;
 	unsigned *val = (unsigned *)STARPU_VECTOR_GET_PTR(descr[0]);
 	val[0]++;
 }
 
-void cpu_codelet_incC(void *descr[], __attribute__ ((unused)) void *_args)
+void cpu_codelet_incC(void *descr[], void *arg)
 {
-	STARPU_SKIP_IF_VALGRIND;
-
+	(void)arg;
 	unsigned *val = (unsigned *)STARPU_VECTOR_GET_PTR(descr[0]);
 	val[2]++;
 }
@@ -84,17 +81,16 @@ void cpu_codelet_incC(void *descr[], __attribute__ ((unused)) void *_args)
 /* increment a = v[0] */
 static struct starpu_codelet cl_inc_a =
 {
-	.where = STARPU_CPU|STARPU_CUDA|STARPU_OPENCL|STARPU_GORDON,
-	.cpu_funcs = {cpu_codelet_incA, NULL},
+	.cpu_funcs = {cpu_codelet_incA},
 #ifdef STARPU_USE_CUDA
-	.cuda_funcs = {cuda_codelet_incA, NULL},
+	.cuda_funcs = {cuda_codelet_incA},
+	.cuda_flags = {STARPU_CUDA_ASYNC},
 #endif
 #ifdef STARPU_USE_OPENCL
-	.opencl_funcs = {opencl_codelet_incA, NULL},
-#endif
-#ifdef STARPU_USE_GORDON
-	.gordon_func = kernel_incA_id,
+	.opencl_funcs = {opencl_codelet_incA},
+	.opencl_flags = {STARPU_OPENCL_ASYNC},
 #endif
+	.cpu_funcs_name = {"cpu_codelet_incA"},
 	.nbuffers = 1,
 	.modes = {STARPU_RW}
 };
@@ -102,17 +98,16 @@ static struct starpu_codelet cl_inc_a =
 /* increment c = v[2] */
 struct starpu_codelet cl_inc_c =
 {
-	.where = STARPU_CPU|STARPU_CUDA|STARPU_OPENCL|STARPU_GORDON,
-	.cpu_funcs = {cpu_codelet_incC, NULL},
+	.cpu_funcs = {cpu_codelet_incC},
 #ifdef STARPU_USE_CUDA
-	.cuda_funcs = {cuda_codelet_incC, NULL},
+	.cuda_funcs = {cuda_codelet_incC},
+	.cuda_flags = {STARPU_CUDA_ASYNC},
 #endif
 #ifdef STARPU_USE_OPENCL
-	.opencl_funcs = {opencl_codelet_incC, NULL},
-#endif
-#ifdef STARPU_USE_GORDON
-	.gordon_func = kernel_incC_id,
+	.opencl_funcs = {opencl_codelet_incC},
+	.opencl_flags = {STARPU_OPENCL_ASYNC},
 #endif
+	.cpu_funcs_name = {"cpu_codelet_incC"},
 	.nbuffers = 1,
 	.modes = {STARPU_RW}
 };
@@ -123,37 +118,25 @@ int main(int argc, char **argv)
 
 #ifdef STARPU_QUICK_CHECK
 	n /= 10;
+#endif
+#ifndef STARPU_LONG_CHECK
 	k /= 8;
 #endif
 
-	ret = starpu_init(NULL);
+	ret = starpu_initialize(NULL, &argc, &argv);
 	if (ret == -ENODEV) return STARPU_TEST_SKIPPED;
 	STARPU_CHECK_RETURN_VALUE(ret, "starpu_init");
 
-#ifdef STARPU_USE_GORDON
-	unsigned elf_id = gordon_register_elf_plugin("./datawizard/sync_and_notify_data_gordon_kernels.spuelf");
-	gordon_load_plugin_on_all_spu(elf_id);
-
-	unsigned kernel_incA_id = gordon_register_kernel(elf_id, "incA");
-	gordon_load_kernel_on_all_spu(kernel_incA_id);
-
-	unsigned kernel_incC_id = gordon_register_kernel(elf_id, "incC");
-	gordon_load_kernel_on_all_spu(kernel_incC_id);
-
-	FPRINTF(stderr, "kernel incA %d incC %d elf %d\n", kernel_incA_id, kernel_incC_id, elf_id);
-#endif
-
 #ifdef STARPU_USE_OPENCL
         ret = starpu_opencl_load_opencl_from_file("tests/datawizard/sync_and_notify_data_opencl_codelet.cl", &opencl_code, NULL);
 	STARPU_CHECK_RETURN_VALUE(ret, "starpu_opencl_load_opencl_from_file");
 #endif
 
-        starpu_vector_data_register(&v_handle, 0, (uintptr_t)v, VECTORSIZE, sizeof(unsigned));
+        starpu_vector_data_register(&v_handle, STARPU_MAIN_RAM, (uintptr_t)v, VECTORSIZE, sizeof(unsigned));
 
 	unsigned iter;
 	for (iter = 0; iter < k; iter++)
 	{
-		int ret;
 		unsigned ind;
 		for (ind = 0; ind < n; ind++)
 		{
@@ -203,7 +186,7 @@ int main(int argc, char **argv)
 		FPRINTF(stderr, "Incorrect result\n");
 		ret = EXIT_FAILURE;
 	}
-	STARPU_RETURN(ret);
+	return ret;
 
 enodev:
 	starpu_data_unregister(v_handle);
